import argparse
import collections
from datetime import date
import gzip
import sys



###################
def revCompIterative(watson): #Gets Reverse Complement
    complements = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 'N': 'N',
                   'R': 'Y', 'Y': 'R', 'S': 'S', 'W': 'W', 'K': 'M',
                   'M': 'K', 'V': 'B', 'B': 'V', 'H': 'D', 'D': 'H'}
    watson = watson.upper()
    watsonrev = watson[::-1]
    crick = ""
    for nt in watsonrev: # make dict to catch bad nts - if more than 1 output them to std error
        try:
            crick += complements[nt]
        except KeyError:
            crick += nt
    return crick
############################

#Output FASTA and GFF separately using the same out_filename but with respective extensions - gz output optional
def write_fasta(UnStORFed_Regions, options):
    if options.out_file == None:
        options.out_file = options.gff.split('.gff')[0]
    if options.gz == False:
        out = open(options.out_file + '_UnStORFed.fasta','w', newline='\n', encoding='utf-8')
    elif options.gz == True:
        out = gzip.open(options.out_file + '_UnStORFed.fasta.gz', 'wt', newline='\n', encoding='utf-8')

    for pos, seq in UnStORFed_Regions.items():
        out.write('>UnStORFed_' + pos + '\n' + seq + '\n')
    out.close()

def write_gff(UnStORFed_Regions,options):
    if options.out_file == None:
        options.out_file = options.gff.split('.gff')[0]
    if options.gz == False:
        out =  open(options.out_file + '_UnStORFed.gff','w', newline='\n', encoding='utf-8')
    elif options.gz == True:
        out = gzip.open(options.out_file + '_UnStORFed.gff.gz', 'wt', newline='\n', encoding='utf-8')
    out.write("##gff-version\t3\n#\tUnStORFed Regions \n#\tRun Date:" + str(date.today()) + '\n')
    out.write("##Original Files: " + options.fasta + ' | ' + options.gff + '\n')
    for pos, seq in UnStORFed_Regions.items():
        length = len(seq)
        pos = pos.replace('_',' ')
        entry = ('Genome\tUnStORFed\tunannotated_region\t' + pos + '\t.\t.\t.\tID=UnStORFed_Region_'+pos+';Note=Length:' + str(length) + '\n')
        out.write(entry)
    out.close()

def fasta_load(fasta_in):
    dna_regions = collections.OrderedDict()
    first = True
    #### Default for when presented with standard fasta file
    for line in fasta_in:
        line = line.strip()
        if line.startswith('#'):
            continue
        elif line.startswith('>') and first == False:  # Check if first seq in file
            dna_regions.update({dna_region_id: [seq, []]})
            seq = ''
            dna_region_id = line.split()[0].replace('>', '').split('_UR_')[1]
        elif line.startswith('>'):
            seq = ''
            dna_region_id = line.split()[0].replace('>', '').split('_UR_')[1]
        else:
            seq += str(line)
            first = False
    dna_regions.update({dna_region_id: [seq, []]})
    return dna_regions

def gff_load(options,gff_in,dna_regions):
    #Will code in different versions for different types of GFF3 files (Prodigal,Ensembl etc)
    for line in gff_in:  # Get gene loci from GFF - ID=Gene will also classify Pseudogenes as genes
        line_data = line.split()
        if line.startswith('\n') or line.startswith('#'): # Not to crash on empty lines in GFF
            continue
        elif line_data[1] == 'StORF_Reporter':
            if 'Con-StORF' not in line:
                current_UR = line_data[8].split('_UR_')[1].split('_StORF')[0]
            else:
                current_UR = line_data[8].split('_UR_')[1].split('_Con-StORF')[0]
            current_UR_StORF = line_data[3] + '_' + line_data[4]
            current_UR_StORF_Frame = line_data[6]
            dna_regions[current_UR][1].append([current_UR_StORF,current_UR_StORF_Frame])
## append storf info to this dict and then go to another func where if theres one storf we just extract normally, if more than one, we likely have to use numpy set info
    return dna_regions

def extractor(options):
    try:
        try: # Detect whether fasta/gff files are .gz or text and read accordingly
            fasta_in = gzip.open(options.fasta,'rt')
            dna_regions = fasta_load(fasta_in)
        except:
            fasta_in = open(options.fasta,'r')
            dna_regions = fasta_load(fasta_in)
        try:
            gff_in = gzip.open(options.gff,'rt')
            dna_regions = gff_load(options,gff_in,dna_regions)
        except:
            gff_in = open(options.gff,'r')
            dna_regions = gff_load(options,gff_in,dna_regions)
    except AttributeError:
        sys.exit("Something Happened")

    UnStORFed_Regions = collections.defaultdict(str)

    for (key,(seq,StORFs))  in dna_regions.items(): #Extract URs from 1 dna_region at a time
        UR_start = int(key.split('_')[0])
        UR_stop = int(key.split('_')[1])
        UR_length = len(seq)
        if len(StORFs) == 1:
            StORF_start = int(StORFs[0][0].split('_')[0])
            StORF_stop = int(StORFs[0][0].split('_')[1])
            relative_StORF_distance_from_start = StORF_start - UR_start
            relative_StORF_distance_from_end = UR_stop - StORF_stop
            if relative_StORF_distance_from_start >= options.minlen:
                UnStORFed_pos = str(UR_start) + '_' + str(StORF_start)
                UnStORFed_start_seq = seq[0:relative_StORF_distance_from_start-1] # -1 is to remove the first bp of the first stop codon of the StORF
                UnStORFed_Regions[UnStORFed_pos] = UnStORFed_start_seq
            if relative_StORF_distance_from_end >= options.minlen:
                UnStORFed_pos = str(StORF_stop) + '_' + str(UR_stop)
                UnStORFed_stop_seq = seq[UR_length-relative_StORF_distance_from_end:]
                UnStORFed_Regions[UnStORFed_pos] = UnStORFed_stop_seq

        elif len(StORFs) > 2:
            prev_StORF_end = 0
            for idx, StORF in enumerate(StORFs):
                StORF_start = int(StORF[0].split('_')[0])
                StORF_stop = int(StORF[0].split('_')[1])

                try:  # Either use UR boundries or the next StORFs boundries
                    next_StORF_start = int(StORFs[idx + 1][0].split('_')[0])
                    relative_StORF_distance_from_start = max(StORF_start - prev_StORF_end, 0)
                    relative_StORF_distance_from_end = max(next_StORF_start - StORF_stop, 0)
                    segment_end = next_StORF_start
                    seq_segment_end = next_StORF_start - UR_start
                except IndexError:  # Next StORF is last StORF
                    relative_StORF_distance_from_start = max(StORF_start - prev_StORF_end, 0)
                    relative_StORF_distance_from_end = UR_stop - StORF_stop
                    segment_end = UR_stop
                    seq_segment_end = -1
                if idx == 0: # sloppy and overwriting from above but its needed for now
                    relative_StORF_distance_from_start = StORF_start - UR_start
                    relative_StORF_distance_from_end = max(next_StORF_start - StORF_stop, 0)
                    segment_start = UR_start
                    seq_segment_start = 0

                else:
                    segment_start = prev_StORF_end
                    seq_segment_start = prev_StORF_end - UR_start


                if relative_StORF_distance_from_start >= options.minlen:
                    UnStORFed_pos = str(segment_start) + '_' + str(StORF_start)
                    UnStORFed_start_seq = seq[seq_segment_start:seq_segment_start+relative_StORF_distance_from_start - 1]  # -1 is to remove the first bp of the first stop codon of the StORF
                    UnStORFed_Regions[UnStORFed_pos] = UnStORFed_start_seq
                if relative_StORF_distance_from_end >= options.minlen:
                    UnStORFed_pos = str(StORF_stop) + '_' + str(segment_end)
                    UnStORFed_stop_seq = seq[seq_segment_end - relative_StORF_distance_from_end:seq_segment_end]
                    UnStORFed_Regions[UnStORFed_pos] = UnStORFed_stop_seq

                prev_StORF_end = StORF_stop

    write_fasta(UnStORFed_Regions, options)
    write_gff(UnStORFed_Regions, options)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='StORF_Reporter v0.5.4: UR_StORFed Run Parameters.')
    parser.add_argument('-f', '--fasta_seq', action='store', dest='fasta', required=True,
                        help='FASTA file for Unannotated Region seq extraction')
    parser.add_argument('-gff', action='store', dest='gff', help='GFF annotation file for the FASTA',
                        required=True)
    parser.add_argument('-min_len', action='store', dest='minlen', default='50', type=int,
                        help='Minimum Un-StORFed UR Length: Default 50')
    parser.add_argument('-o', '--output_file', action='store', dest='out_file', required=False,
                        help='Output file - Without filetype - default appends "_UnStORFed.fasta/.gff" to end of input gff filename')
    parser.add_argument('-gz', action='store', dest='gz', default='False', type=eval, choices=[True, False],
                        help='Default - False: Output as .gz')
    parser.add_argument('-v', action='store', dest='verbose', default='False', type=eval, choices=[True, False],
                        help='Default - False: Print out runtime status')

    options = parser.parse_args()
    extractor(options)

    # Contig name could have a ';' which will mess up later on in StORF_Reporter-R
    # UR output should state original non extended

    ## The current version reports UnStORFed regions according to the extracted URs - So the


import argparse


try:
    from ORForise.GFF_Adder import gff_adder  # Calling from ORForise via pip
except (ModuleNotFoundError, ImportError, NameError, TypeError) as error:
    from ORForise.src.ORForise.GFF_Adder import gff_adder  # Calling from ORForise locally (StORF_Reporter and ORForise in same dir)





if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='StORF-Reporter v0.5.4: StORF-Adder Run Parameters.')
    parser.add_argument('-dna', '--genome_DNA', required=True, help='Genome DNA file (.fa) which both annotations '
                                                                    'are based on')
    parser.add_argument('-rt', '--reference_tool', required=False,
                        help='Which tool format to use as reference? - If not provided, will default to '
                             'standard Ensembl GFF format, can be Prodigal or any of the other tools available')
    parser.add_argument('-ref', '--reference_annotation', required=True,
                        help='Which reference annotation file to use as reference?')
    parser.add_argument('-gi', '--gene_ident',  default='CDS', required=False,
                        help='Identifier used for extraction of "genic" regions from reference annotation '
                             '"CDS,rRNA,tRNA": Default for is "CDS"')
    parser.add_argument('-at', '--additional_tool', default='StORF_Reporter', required=False, # needs cleaner fix
                        help='Which format to use for additional annotation?')
    parser.add_argument('-add', '--additional_annotation', required=True,
                        help='Which annotation file to add to reference annotation?')
    parser.add_argument('-olap', '--overlap', default=50, type=int, required=False,
                        help='Maximum overlap between reference and additional genic regions (CDS,rRNA etc) - Default: 50 nt')
    parser.add_argument('-o', '--output_file', required=True,
                        help='Output filename')

    options = parser.parse_args()
    gff_adder(options)



import argparse
import collections
from datetime import date
import gzip
import sys


try:
    from .Constants import *
except (ModuleNotFoundError, ImportError, NameError, TypeError) as error:
    from Constants import *

#Output FASTA and GFF separately using the same filename but with respective extensions - gz output optional
def write_fasta(dna_regions, options, fasta_out):
    fasta_out.write("##\tUR-Extractor \n#\tRun Date:" + str(date.today()) + '\n')
    fasta_out.write('##StORF-Reporter ' + StORF_Reporter_Version + '\n')
    fasta_out.write("##Original Files: " + options.fasta.split('/')[-1] + ' | ' + options.gff.split('/')[-1] + '\n')
    for dna_region, dna_region_ur in dna_regions.items():
        fasta_out.write('\n##sequence-region\t' + dna_region + ' 1 ' + str(len(dna_region_ur[0])) + '\n')
        ur_ident = dna_region + options.ident # Add user ident onto name of dna regions
        if dna_region_ur[3]:
            for ex_ur, data in dna_region_ur[3].items():
                original_ur = data[0]
                ur_seq = data[1]
                fasta_out.write('>' + ur_ident + '_' + ex_ur + '\n' + ur_seq + '\n')
    fasta_out.close()

def write_gff(dna_regions,options,gff_out):
    gff_out.write("##gff-version\t3\n#\tUR-Extractor \n#\tRun Date:" + str(date.today()) + '\n')
    gff_out.write('##StORF-Reporter ' + StORF_Reporter_Version + '\n')
    for seq_reg in dna_regions:
        gff_out.write('##sequence-region\t' + seq_reg + ' 1 ' + str(dna_regions[seq_reg][1]) + '\n')
    gff_out.write("##Original Files: " + options.fasta.split('/')[-1] + ' | ' + options.gff.split('/')[-1] + '\n\n')
    for dna_region, dna_region_ur in dna_regions.items():
        ur_ident = dna_region + options.ident
        if dna_region_ur[3]:
            for ex_ur, data in dna_region_ur[3].items():
                length = len(data[1])
                ex_ur_pos = ex_ur.replace('_','\t')
                entry = (dna_region + '\tUR_Extractor\tunannotated_region\t' + ex_ur_pos + '\t.\t.\t.\tID='+ur_ident+'_'+ex_ur+';' + 'Original_UR=' + str(data[0]) + ';Note=UR_Length(Extended):' + str(length) + '\n')
                gff_out.write(entry)
    gff_out.close()

def fasta_load(fasta_in):
    dna_regions = collections.OrderedDict()
    first = True
    if '>' in fasta_in.readline().rstrip():
        fasta_in.seek(0)
        #### Default for when presented with standard fasta file
        for line in fasta_in:
            line = line.strip()
            if line.startswith('>') and first == False:  # Check if first seq in file
                dna_region_length = len(seq)
                dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
                seq = ''
                dna_region_id = line.split()[0].replace('>', '')
            elif line.startswith('>'):
                seq = ''
                dna_region_id = line.split()[0].replace('>', '')
            else:
                seq += str(line)
                first = False
        dna_region_length = len(seq)
        dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
    elif '##' in  fasta_in.readline().rstrip(): # Clunky and may fall over
        fasta_in.seek(0)
        #### Called when presented with Prokka GFF file so must get fasta from inside it
        ### Get to genome seq
        at_FASTA = False
        for line in fasta_in:  # Get gene loci from GFF - ID=Gene will also classify Pseudogenes as genes
            if line.startswith('##FASTA'):  # Not to crash on empty lines in GFF
                at_FASTA = True
            elif at_FASTA == True:
                line = line.strip()
                if line.startswith('>') and first == False:  # Check if first seq in file
                    dna_region_length = len(seq)
                    dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
                    seq = ''
                    dna_region_id = line.split()[0].replace('>', '')
                elif line.startswith('>'):
                    seq = ''
                    dna_region_id = line.split()[0].replace('>', '')
                else:
                    seq += str(line)
                    first = False
        dna_region_length = len(seq)
        dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})

    return dna_regions

def pyrodigal_virtual_gff_load(gff_in,dna_regions):
    with open(gff_in.name, "r") as gff:
        tmp = gff.read()
        for line in tmp.splitlines():
            line_data = line.split()
            if line.startswith('\n') or line.startswith('#'):  # Not to crash on empty lines in GFF
                continue
            else:
                try:
                    if line_data[0] in dna_regions:
                        if 'CDS' in line_data[2]: # line[2] for normal run
                            pos = line_data[3] + '_' + line_data[4]
                            if pos not in dna_regions[line_data[0]][2]:
                                dna_regions[line_data[0]][2].append(pos) # This will add to list
                except IndexError:
                    continue
    return dna_regions

def gff_load(options,gff_in,dna_regions):
    for line in gff_in:  # Get gene loci from GFF - ID=Gene will also classify Pseudogenes as genes
        line_data = line.split()
        if line.startswith('\n') or line.startswith('#'):  # Not to crash on empty lines in GFF
            continue
        elif options.gene_ident[0] == 'ID=gene':
            if line_data[0] in dna_regions and options.gene_ident[0] in line_data[8]:
                pos = line_data[3] + '_' + line_data[4]
                dna_regions[line_data[0]][2].append(pos) # This will add to list
        else:
            try:
                if line_data[0] in dna_regions:
                    if any(gene_type in line_data[2] for gene_type in options.gene_ident): # line[2] for normal run
                        pos = line_data[3] + '_' + line_data[4]
                        if pos not in dna_regions[line_data[0]][2]:
                            dna_regions[line_data[0]][2].append(pos) # This will add to list
            except IndexError:
                continue
    return dna_regions

def extractor(options):
    if options.annotation_type[0] == 'Pyrodigal':
        try:  # Detect whether fasta/gff files are .gz or text and read accordingly
            fasta_in = gzip.open(options.fasta, 'rt')
            dna_regions = fasta_load(fasta_in)
        except:
            fasta_in = open(options.fasta, 'r', encoding='unicode_escape')
            dna_regions = fasta_load(fasta_in)
        dna_regions = pyrodigal_virtual_gff_load(options.gff, dna_regions)
    else:
        try:
            try: # Detect whether fasta/gff files are .gz or text and read accordingly
                fasta_in = gzip.open(options.fasta,'rt')
                dna_regions = fasta_load(fasta_in)
            except:
                fasta_in = open(options.fasta,'r',encoding='unicode_escape')
                dna_regions = fasta_load(fasta_in)
            try:
                gff_in = gzip.open(options.gff,'rt')
                dna_regions = gff_load(options,gff_in,dna_regions)
            except:
                gff_in = open(options.gff,'r',encoding='unicode_escape')
                dna_regions = gff_load(options,gff_in,dna_regions)
        except AttributeError:
            sys.exit("Attribute Error:\nStORF'ed GFF probably already exists - Must be deleted before running")


    for (key,(seq,seq_length,posns,URs))  in dna_regions.items(): #Extract URs from 1 dna_region at a time
        unannotated_regions = collections.OrderedDict()
        unannotated_start = 0
        if posns: # If UR has a pos
            for pos in posns: # Iterate over GFF loci and measure flanking regions for potential URs
                start = int(pos.split('_')[0])
                stop = int(pos.split('_')[1])
                ###### This hack is to get over GFF errors where genome-long annotations
                if stop-start >= 100000:
                    if options.verbose == True:
                        print("UR " + pos + " is more than 100,000 kbs - Please Check Annotation")
                    continue
                if start > unannotated_start:
                    length = start - unannotated_start
                    if length >= options.minlen and length <= options.maxlen: #default between 30 - 100,000
                        original_UR = str(max(unannotated_start,1)) + '_' + str(start)
                        unannotated_start = max(unannotated_start - options.exlen, 0) # 0 here for pythons base-o here is for the GFF base-1 reporting format
                        start += options.exlen
                        unannotated_seq = seq[unannotated_start:start]
                        unannotated_loci = str(max(unannotated_start,1)) + '_' + str(start)
                        unannotated_regions.update({unannotated_loci: [original_UR,unannotated_seq]})
                if stop > unannotated_start:
                    unannotated_start = stop
            try:
                if (seq_length - stop) >= options.minlen and (seq_length - stop) <= options.maxlen: #default between 30 - 100,000: # Get UR at end of dna_region if longer than minlen
                    original_UR = str(max(unannotated_start, 1)) + '_' + str(start)
                    stop -= options.exlen
                    unannotated_seq = seq[stop:seq_length]
                    unannotated_loci = str(stop) + '_' + str(seq_length)
                    unannotated_regions.update({unannotated_loci: [original_UR,unannotated_seq]})
            except UnboundLocalError:
                pass
            dna_regions.update({key: (seq, seq_length, posns, unannotated_regions)})


    if options.nout == False:
        write_fasta(dna_regions, options, options.fasta_out)
        write_gff(dna_regions, options, options.gff_out)
    else:
        return dna_regions


def main():
    print("Thank you for using StORF-Reporter -- A detailed user manual can be found at https://github.com/NickJD/StORF-Reporter\nPlease report any issues to: https://github.com/NickJD/StORF-Reporter/issues\n#####")

    parser = argparse.ArgumentParser(description='StORF-Reporter ' + StORF_Reporter_Version + ': UR-Extractor Run Parameters.')
    parser._action_groups.pop()

    required = parser.add_argument_group('Required Arguments')
    required.add_argument('-f', action='store', dest='fasta', required=True,
                        help='FASTA file for Unannotated Region seq extraction')
    required.add_argument('-gff', action='store', dest='gff', help='GFF annotation file for the FASTA',
                        required=True)

    optional = parser.add_argument_group('Optional Arguments')
    optional.add_argument('-ident', action='store', dest='ident', default='_UR',
                        help='Identifier given for Unannotated Region output sequences - Do not modify if output is '
                             'to be used by StORF-Finder: Default "Sequence-ID"_UR')
    optional.add_argument('-min_len', action='store', dest='minlen', default='30', type=int,
                        help='Minimum UR Length: Default 30')
    optional.add_argument('-max_len', action='store', dest='maxlen', default='100000', type=int,
                        help='Maximum UR Length: Default 100,000')
    optional.add_argument('-ex_len', action='store', dest='exlen', default='50', type=int,
                        help='UR Extension Length on 5\' and 3\': Default 50')
    optional.add_argument('-gene_ident', action='store', dest='gene_ident', default='ID=gene',
                        help='Identifier used for extraction of Unannotated Regions "CDS,rRNA,tRNA": Default for Ensembl_Bacteria = '
                             '"ID=gene" or "-gene_ident CDS" for "most" genome annotations')

    output = parser.add_argument_group('Output')
    output.add_argument('-oname', action="store", dest='o_name', required=False,
                        help='Default - Appends \'_UR\' to end of input FASTA filename')
    output.add_argument('-odir', action="store", dest='o_dir', required=False,
                        help='Default -  Same directory as input FASTA')
    output.add_argument('-gz', action='store', dest='gz', default='False', type=eval, choices=[True, False],
                        help='Default - False: Output as .gz')


    misc = parser.add_argument_group('Misc')
    misc.add_argument('-v', action='store', dest='verbose', default='False', type=eval, choices=[True, False],
                        help='Default - False: Print out runtime status')
    parser.add_argument('-nout', action='store', dest='nout', default='False', type=eval, choices=[True, False],
                        help=argparse.SUPPRESS)
    parser.add_argument('-pyrodigal', action='store', dest='pyrodigal', default='False', type=eval, choices=[True, False],
                        help=argparse.SUPPRESS)
    parser.add_argument('-nout_pyrodigal', action='store', dest='nout_pyrodigal', default='True', type=eval, choices=[True, False],
                        help=argparse.SUPPRESS)


    options = parser.parse_args()
    options.annotation_type = [None,None]

    #### Output Directory and Filename handling
    if options.o_dir == None and options.o_name == None:
        tmp_extension = options.fasta.split('.')[-1]  # could be .fa/.fasta etc
        output_file = options.fasta.replace('.' + tmp_extension, '')
        output_file = output_file + '_UR'
    elif options.o_dir != None and options.o_name != None:
        output_file = options.o_dir
        output_file = output_file + '/' if not output_file.endswith('/') else output_file
        output_file = output_file + options.o_name
    elif options.o_dir != None:
        tmp_extension = options.fasta.split('.')[-1]  # could be .fa/.fasta etc
        output_file = options.fasta.replace('.' + tmp_extension, '').split('/')[-1]
        output_file = options.o_dir + output_file + '_UR'
    elif options.o_name != None:
        tmp_filename = options.fasta.split('/')[-1]  # could be .fa/.fasta etc
        output_file = options.fasta.replace(tmp_filename, '')
        output_file = output_file + options.o_name

    if options.gz == False:
        options.fasta_out = open(output_file + '.fasta','w', newline='\n', encoding='utf-8')
        options.gff_out =  open(output_file + '.gff','w', newline='\n', encoding='utf-8')
    elif options.gz == True:
        options.fasta_out = open(output_file + '.fasta.gz','wt', newline='\n', encoding='utf-8')
        options.gff_out =  open(output_file + '.gff.gz','wt', newline='\n', encoding='utf-8')

    options.gene_ident = options.gene_ident.split(',')
    extractor(options)


if __name__ == "__main__":
    main()
    print("Complete")







import argparse
from argparse import Namespace
import pathlib
import collections
import textwrap
import hashlib
import pyrodigal
from tempfile import NamedTemporaryFile
import os
import sys
import gzip


try:
    from .UR_Extractor import extractor
    from .StORF_Finder import StORF_Reported
    from .Constants import *
except (ModuleNotFoundError, ImportError, NameError, TypeError) as error:
    from UR_Extractor import extractor
    from StORF_Finder import StORF_Reported
    from Constants import *


class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        return argparse.HelpFormatter._split_lines(self, text, width)

# class StORF_obj: # TODO
#     def __init__(self, name, roll):
#         self.name = name
#         self.roll = roll

###################
gencode = {
      'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M',
      'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T',
      'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K',
      'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R',
      'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L',
      'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P',
      'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q',
      'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R',
      'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V',
      'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A',
      'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E',
      'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G',
      'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S',
      'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L',
      'TAC':'Y', 'TAT':'Y', 'TAA':'*', 'TAG':'*',
      'TGC':'C', 'TGT':'C', 'TGA':'*', 'TGG':'W'}

def translate_frame(sequence):
    translate = ''.join([gencode.get(sequence[3 * i:3 * i + 3], 'X') for i in range(len(sequence) // 3)])
    return translate
############################

def GFF_StoRF_write(StORF_options, Reporter_options, track_contig, gff_out, StORF, StORF_Num): # Consistency in outfile
    ID = track_contig + '_UR_' + StORF[0] + '_' + StORF[10] + '_'+ str(StORF[9])
    try:
        to_hash = gff_out.split('.')[0] + '_' + ID # create unique hash from outfile name and ID
    except AttributeError:
        to_hash = gff_out.name.split('.')[0] + '_' + ID  # create unique hash from outfile name and ID
    locus_tag = hashlib.shake_256(to_hash.encode()).hexdigest(8)
    ### Write out new GFF entry -
    strand = StORF[7]
    start = StORF[3]
    stop = StORF[4]
    UR_Start = int(StORF[0].split('_')[0])
    start_stop = StORF[12][0:3]
    if len(StORF[11]) == 3: # fix for con-storfs
        mid = int(StORF[11]) - int(StORF[11])
        mid_stop = StORF[12][mid - 3:mid]
    else:
        mid_stop = 'N/A'
    end_stop = StORF[12][-3:]

    if strand == '+':
        gff_start = max(start + 1 + UR_Start, 1) # this may need to change to account for different exlen
        gff_stop = max(stop + UR_Start, 1)
        if UR_Start == 1:
            gff_start -= 1
            gff_stop -= 1

        if Reporter_options.stop_inclusive == False:  # To remove the start and stop codon positions.
            gff_start = gff_start + 3
        frame = (int(gff_stop) % 3) + 1
    elif strand == '-':
        gff_start = max(start - 2 + UR_Start, 1)
        gff_stop = max(stop - 3 + UR_Start, 1)
        if Reporter_options.stop_inclusive == False:  # To remove the start and stop codon positions.
            gff_stop = gff_stop - 3
        frame = (int(gff_stop) % 3) + 4

    StORF_length = int(gff_stop) + 1 - int(gff_start) # +1 to adjust for base-1

    gff_out.write(track_contig + '\tStORF-Reporter\t' + StORF_options.feature_type + '\t' +  str(gff_start) + '\t' + str(gff_stop) + '\t.\t' +
        StORF[7] + '\t0\tID=StORF_' + locus_tag + ';locus_tag=' + ID + ';INFO=Additional_Annotation_StORF-Reporter;UR_Stop_Locations=' + StORF[11].replace(',','-') + ';Name=' +
           StORF[10] + '_' + str(StORF_Num) + ';' + StORF[10] + '_Num_In_UR=' + str(StORF[9]) + ';' + StORF[10] + '_Length=' + str(StORF_length) + ';' + StORF[10] +
           '_Frame=' + str(frame) + ';UR_' + StORF[10] + '_Frame=' + str(StORF[6]) +  ';Start_Stop=' + start_stop + ';Mid_Stops=' + mid_stop  + ';End_Stop='
           + end_stop + ';StORF_Type=' + StORF[10] + '\n')


def FASTA_StORF_write(Reporter_options, track_contig, fasta_out, StORF):  # Consistency in outfile
    ## This should compute the same hash as the one for GFF_Write - Should not be computed twice though
    ID = track_contig + '_UR_' + StORF[0] + '_' + StORF[10] + '_'+ str(StORF[9])
    to_hash = fasta_out.name.split('.')[0] + '_' + ID # create unique hash from outfile name and ID
    locus_tag = 'StORF_' + hashlib.shake_256(to_hash.encode()).hexdigest(8)
    ### Wrtie out new FASTA entry - Currently only write out as nt
    fasta_out.write('>'+locus_tag+'\n')
    sequence = StORF[-1]
    if Reporter_options.translate == True:
        sequence = translate_frame(sequence[0:])
        if Reporter_options.remove_stop == True:
            sequence = sequence.strip('*')
    if Reporter_options.line_wrap == True:
        wrapped = textwrap.wrap(sequence, width=60)
        for wrap in wrapped:
            fasta_out.write(wrap + '\n')
    else:
        fasta_out.write(sequence+'\n')
    if Reporter_options.storfs_out == True: # Not efficient
        if Reporter_options.gz == False:
            storf_fasta_outfile = open(fasta_out.name.replace('.fasta','_StORFs_Only.fasta'),'a')
        else:
            storf_fasta_outfile = open(fasta_out.name.replace('.fasta.gz','_StORFs_Only.fasta.gz'),'a')
        storf_fasta_outfile.write('>' + locus_tag + '\n')
        if Reporter_options.line_wrap == True:
            wrapped = textwrap.wrap(sequence, width=60)
            for wrap in wrapped:
                storf_fasta_outfile.write(wrap+'\n')
        else:
            storf_fasta_outfile.write(sequence+'\n')

def FASTA_Load(faa_infile,ffn_infile):
    ### Coding sequences first
    Prokka_AA = collections.defaultdict()
    first = True
    infile = open(faa_infile)
    seq = ''
    ##load in fasta file and make dict of seqs and ids
    for line in infile:
        line = line.strip()
        if line.startswith('>') and first == False:
            Prokka_AA.update({id:seq})
            id = line.replace('>','')
            id = id.split(' ')[0]
            seq = ''
        elif line.startswith('>'):
            id = line.replace('>','')
            id = id.split(' ')[0]
            first = False
        else:
            seq += line
    Prokka_AA.update({id:seq})
    ### Coding and Non-Coding sequences in NT form next
    Prokka_NT = collections.defaultdict()
    first = True
    infile = open(ffn_infile)
    seq = ''
    ##load in fasta file and make dict of seqs and ids
    for line in infile:
        line = line.strip()
        if line.startswith('>') and first == False:
            Prokka_NT.update({id:seq})
            id = line.replace('>','')
            id = id.split(' ')[0]
            seq = ''
        elif line.startswith('>'):
            id = line.replace('>','')
            id = id.split(' ')[0]
            first = False
        else:
            seq += line
    Prokka_NT.update({id:seq})
    return Prokka_AA,Prokka_NT

def read_pyrodigal_fasta(fasta_in,sequences):
    first = True
    for line in fasta_in:
        line = line.strip()
        if line.startswith(';'):
            continue
        elif line.startswith('>') and not first:
            sequences.update({sequence_name: seq})
            seq = ''
            sequence_name = line.split(' ')[0].replace('>', '')
        elif line.startswith('>'):
            seq = ''
            sequence_name = line.split(' ')[0].replace('>', '')
        else:
            seq += str(line)
            first = False
        sequences.update({sequence_name: seq})
    return sequences

def pyrodigal_predict(fasta,Reporter_options):
    sequences = collections.OrderedDict()
    try:  # Detect whether fasta/gff files are .gz or text and read accordingly
        fasta_in = gzip.open(fasta, 'rt')
        sequences = read_pyrodigal_fasta(fasta_in, sequences)
    except:
        fasta_in = open(fasta, 'r', encoding='unicode_escape')  # Not sure if needed in long term
        sequences = read_pyrodigal_fasta(fasta_in, sequences)
    pyrodigal_hold = NamedTemporaryFile(mode='w+', delete=False)

    orf_finder = pyrodigal.OrfFinder()
    longest = (max(sequences.values(), key=len)) # Get longest contig for training

    if Reporter_options.py_train == 'meta':
        orf_finder = pyrodigal.OrfFinder(meta=True)
    elif Reporter_options.py_train == 'longest':
        orf_finder.train(longest)
    for sequence_name, seq in sequences.items():
        if Reporter_options.py_train == 'indvidual':
            try:  # Catch if sequence is less than 20kb - will run in meta mode
                orf_finder.train(seq)
                break
            except ValueError:
                orf_finder = pyrodigal.OrfFinder(meta=True)  # If NO contig is >= 20kb#
        else:
            genes = orf_finder.find_genes(seq)
        if Reporter_options.py_unstorfed == True:
            genes.write_genes(Reporter_options.pyrodigal_out_file_fasta, sequence_id=sequence_name)
            genes.write_gff(Reporter_options.pyrodigal_out_file_gff, sequence_id=sequence_name)
        genes.write_gff(pyrodigal_hold, sequence_id=sequence_name)
    if Reporter_options.py_unstorfed == True:
        Reporter_options.pyrodigal_out_file_fasta.close()
        Reporter_options.pyrodigal_out_file_gff.close()
    pyrodigal_hold.close()
    Reporter_options.gff = pyrodigal_hold
    Reporter_options.fasta = fasta
    URs = extractor(Reporter_options)
    return URs, Reporter_options


def run_UR_Extractor_Directory(Reporter_options): # When given a complete Prokka/Bakta Directory
    identifier = Reporter_options.path.split('/')[-2]
    Reporter_options.fasta = Reporter_options.path + identifier + '.fna'
    if Reporter_options.annotation_type[0] == 'Prokka':
        Reporter_options.gff = Reporter_options.fasta.replace('.fna','.gff')
    if Reporter_options.annotation_type[0] == 'Bakta':
        Reporter_options.gff = Reporter_options.fasta.replace('.fna', '.gff3')
    URs = extractor(Reporter_options)
    return URs,Reporter_options


def run_UR_Extractor_Extended_GFFs(Reporter_options,gff): # When given a directory with multiple GFFs but without accompianing .fna
    if '_StORF-Reporter_Extended' not in str(gff): #Might fall over - put a break
        gff = str(gff)
        Reporter_options.gff = gff
    Reporter_options.fasta = gff
    URs = extractor(Reporter_options)
    return URs,Reporter_options

def run_UR_Extractor_Matched(Reporter_options,gff): # When given a directory with multiple GFFs but with accompianing .fna
    if '_StORF-Reporter_Extended' not in str(gff): #Might fall over - put a break
        gff = str(gff)
        Reporter_options.gff = gff
        fasta = gff.replace('.gff','.fasta')
        fna = gff.replace('.gff', '.fna')
        fa = gff.replace('.gff', '.fa')
        if os.path.isfile(fasta):
            Reporter_options.fasta = fasta
        elif os.path.isfile(fna):
            Reporter_options.fasta = fna
        elif os.path.isfile(fa):
            Reporter_options.fasta = fa
        else:
            sys.exit('No matching FASTA file found for ' + gff)
    URs = extractor(Reporter_options)
    return URs,Reporter_options

def run_UR_Extractor_GFF(Reporter_options, fasta, gff): # When given a directory with multiple GFFs but without accompianing .fna
    if '_StORF-Reporter_Extended' not in gff: #Might fall over - put a break
        gff = gff
        Reporter_options.gff = gff
    fasta = str(fasta)
    Reporter_options.fasta = fasta
    URs = extractor(Reporter_options)
    return URs,Reporter_options

def find_prev_StORFs(StORF_options, Reporter_options, Contig_URs, track_current_start, track_prev_stop, track_contig):
    StORFs_to_del = []
    StORFs = []
    is_break = False
    try:
        current_Contig_URs = Contig_URs[track_contig]
        for StORF_Num, data in current_Contig_URs.items():
            if is_break == True:
                break
            if data != None:
                ur_pos = data[8] # extended UR
                ur_start = int(ur_pos.split('_')[0])
                StORF_Stop_Pos = data[0]
                StORF_Start_In_UR = int(StORF_Stop_Pos.split(',')[0])
                StORF_Stop_In_UR = int(StORF_Stop_Pos.split(',')[-1])
                ur_frame = data[2]
                strand = data[3]
                gff_start = str(StORF_Start_In_UR + 1 + max(ur_start - Reporter_options.exlen, 0))
                gff_stop = str(StORF_Stop_In_UR + max(ur_start - Reporter_options.exlen, 0))
                allow_start = int(gff_start) + StORF_options.allowed_overlap
                allow_stop = int(gff_stop) - StORF_options.allowed_overlap
                StORF_Seq = data[1]
                StORF_Length = data[4]
                StORF_UR_Num = data[6]
                if allow_stop <= track_current_start:# and allow_start >= track_prev_stop:
                    if strand == '+': # Get original genome frame
                        frame = (int(gff_start) % 3) + 1
                    elif strand == '-':
                        frame = (int(gff_stop) % 3) + 4
                    StORFs.append([ur_pos,gff_start,gff_stop,StORF_Start_In_UR,StORF_Stop_In_UR,frame,ur_frame,strand,
                                   StORF_Length,StORF_UR_Num,data[5],StORF_Stop_Pos,StORF_Seq])
                    if StORF_Num not in StORFs_to_del:
                        StORFs_to_del.append(StORF_Num)
                elif allow_start > track_current_start: # Check
                    is_break = True
                    break

        for StORF in StORFs_to_del:
            del current_Contig_URs[StORF]
        Contig_URs[track_contig] = current_Contig_URs # Just incase shallow copy does not remove StORF for us
    ### Might need to force remove more ur storfs regions

        return StORFs, Contig_URs

    except KeyError:
        if StORF_options.verbose == True:
            print("GFF formatting error - Something wrong with seqeunce region " + track_contig)
        return StORFs, Contig_URs


def find_after_StORFs(StORF_options,Contig_URs,track_current_start,track_current_stop, track_contig):
    StORFs = []
    is_break = False
    try:
        current_Contig_URs = Contig_URs[track_contig]
        for StORF_Num, data in current_Contig_URs.items():
            if is_break == True:
                break
            if data != None:
                # for UR_StORF, UR_StORF_data in UR_StORFs.items():
                ur_pos = data[8] # extended UR
                StORF_Stop_Pos = data[0]
                StORF_Start_In_UR = int(StORF_Stop_Pos.split(',')[0])
                StORF_Stop_In_UR = int(StORF_Stop_Pos.split(',')[-1])
                ur_frame = data[2]
                strand = data[3]
                key_start = int(ur_pos.split('_')[0])
                StORF_start = key_start + StORF_Start_In_UR
                StORF_stop = key_start + StORF_Stop_In_UR
                allow_start = StORF_start + StORF_options.allowed_overlap
                StORF_Seq = data[1]
                StORF_Length = data[4]
                StORF_UR_Num = data[6]
            if allow_start >= track_current_stop:
                gff_start = str(StORF_Start_In_UR + int(ur_pos.split('_')[-1].split('_')[0]))
                gff_stop = str(StORF_Stop_In_UR + int(ur_pos.split('_')[-1].split('_')[0]))
                if strand == '+':  # Get original genome frame
                    frame = (int(gff_start) % 3) + 1
                elif strand == '-':
                    frame = (int(gff_stop) % 3) + 4
                StORFs.append(
                    [ur_pos, StORF_start, StORF_stop, StORF_Start_In_UR, StORF_Stop_In_UR, frame, ur_frame, strand,
                     StORF_Length, StORF_UR_Num, data[5], StORF_Stop_Pos, StORF_Seq])

        return StORFs
    except KeyError:
        if StORF_options.verbose == True:
            print("GFF formatting error - Something wrong with sequence region " + track_contig)
        return StORFs

def StORF_Filler(Reporter_options, Reported_StORFs):
    if Reporter_options.annotation_type[0] == 'Pyrodigal':
        with open(Reporter_options.gff.name, "r") as gff:
            gff_in = gff.read()
            gff_name = Reporter_options.path.split('.')[0]
    else:
        with open(Reporter_options.gff, 'r') as gff:
            gff_in = gff.read()
            gff_name = Reporter_options.gff
    try:
        Reporter_options.gff_outfile = open(Reporter_options.output_file + '.gff', 'w')
    except FileNotFoundError:
        sys.exit('File not found error - Try providing full path.')
    track_prev_contig, track_contig = '',''
    StORF_Num = 0
    end = False
    written_line = None
    Contig_URS = collections.defaultdict()
    for Contig, URs in Reported_StORFs.items():
        UR_StORF_Num = 0
        Contig_StORFs = collections.OrderedDict()
        for UR in URs:
            for StORF, data in UR.items():
                data.insert(0, StORF)
                Contig_StORFs.update({UR_StORF_Num: data})
                UR_StORF_Num += 1
        Contig_URS[Contig] = Contig_StORFs
    ### Rename .gff to .fasta / faa and load in fasta file
    if Reporter_options.annotation_type[1] == 'Out_Dir': # If normal Prokka Dir run
        if Reporter_options.gz == False:
                fasta_outfile = open(Reporter_options.output_file + '.fasta', 'w', newline='\n', encoding='utf-8')
        elif Reporter_options.gz == True:
                fasta_outfile = open(Reporter_options.output_file + '.fasta.gz', 'wt', newline='\n', encoding='utf-8')
        faa_infile = Reporter_options.gff.replace('.gff3', '.faa').replace('.gff','.faa')
        ffn_infile = Reporter_options.gff.replace('.gff3', '.ffn').replace('.gff','.ffn')
        Original_AA,Original_NT = FASTA_Load(faa_infile,ffn_infile)
    if Reporter_options.annotation_type[0] == 'Pyrodigal':
        Reporter_options.gff_outfile.write('##Pyrodigal annotation and StORF-Reporter extended GFF annotation of ' + Reporter_options.fasta.split('/')[-1] + '\n')
    else:
        Reporter_options.gff_outfile.write('##StORF-Reporter extended annotation of ' + gff_name.split('/')[-1] + '\n')
    Reporter_options.gff_outfile.write('##StORF-Reporter ' + StORF_Reporter_Version + '\n')

    first_region = True
    for line in gff_in.splitlines( ):
        if not line.startswith('#') and end == False:
            data = line.split('\t')
            track_current_start = int(data[3])
            track_current_stop = int(data[4])
            track_contig = data[0]
            if data[2] == 'region':
                Reporter_options.gff_outfile.write(line.strip() + '\n')
                first_region = False
                continue
            if track_contig != track_prev_contig:  # End of current contig
                # if track_prev_contig != '': # Get last StORF on Contig - If Present
                #     StORFs = find_after_StORFs(StORF_options, Contig_URS, track_prev_start, track_prev_stop,track_prev_contig) # Changed to prev stop because we are switching from previous contig
                #     if StORFs:
                #         for StORF in StORFs:
                #             GFF_StoRF_write(StORF_options, Reporter_options, track_prev_contig, outfile, StORF,
                #                             StORF_Num)  # To keep consistency
                #             if StORF_options.path == True:
                #                 FASTA_StORF_write(Reporter_options, track_contig, fasta_outfile, StORF)
                #             StORF_Num += 1
                track_prev_start, track_prev_stop = 0, 0
            track_prev_contig = track_contig
            if track_current_start == track_prev_start and track_current_stop == track_prev_stop:  # `duplicate' entry in GFF
                tracked = True
            else:
                StORFs, Contig_URS = find_prev_StORFs(Reporter_options, Reporter_options, Contig_URS, track_current_start, track_prev_stop, track_contig)
                tracked = False
            track_prev_start = track_current_start
            track_prev_stop = track_current_stop
            ##### Print out Prokka/Bakta Gene
            if tracked == False and Reporter_options.annotation_type[1] == 'Out_Dir':
                if ('gene' in data[2] or 'ID=' in data[8]) and Reporter_options.annotation_type[0] in ('Prokka', 'Bakta'):
                    Original_ID = data[8].split(';')[0].replace('ID=','').replace('_gene','')
                    try:
                        Original_Seq = Original_NT[Original_ID] # used to be AA but changed for to NT. CHECK
                    except KeyError:
                        try:
                            Original_Seq = Original_NT[Original_ID]
                        except KeyError:
                            if Reporter_options.verbose == True:
                                print("Original seq " + Original_ID + " not found")
                    fasta_outfile.write('>'+Original_ID+'\n')
                    if Reporter_options.translate == True:
                        Original_Seq = translate_frame(Original_Seq[0:])
                        if Reporter_options.remove_stop == True:
                            Original_Seq = Original_Seq.strip('*')
                    if Reporter_options.line_wrap == True:
                        wrapped = textwrap.wrap(Original_Seq, width=60)
                        for wrap in wrapped:
                            fasta_outfile.write(wrap + '\n')
                    else:
                        fasta_outfile.write(Original_Seq+'\n')
            if StORFs:
                for StORF in StORFs:  # ([ur_pos,StORF_start, StORF_stop, StORF_Start_In_UR, StORF_Stop_In_UR, frame, ur_frame, strand, StORF_Length, StORF_UR_Num, StORF_Seq)]
                    GFF_StoRF_write(Reporter_options, Reporter_options, track_contig, Reporter_options.gff_outfile, StORF, StORF_Num)  # To keep consistency
                    if Reporter_options.annotation_type[1] == 'Out_Dir':
                        FASTA_StORF_write(Reporter_options, track_contig, fasta_outfile, StORF)
                    StORF_Num += 1
            if line != written_line:
                Reporter_options.gff_outfile.write(line.strip()+'\n')
                written_line = line
            StORFs = None
        elif line.startswith('##sequence-region') and first_region != True:
            StORFs = find_after_StORFs(Reporter_options, Contig_URS, track_prev_start, track_prev_stop,
                                       track_prev_contig)  # Changed to prev stop because we are switching from previous contig
            if StORFs:
                for StORF in StORFs:
                    GFF_StoRF_write(Reporter_options, track_prev_contig, Reporter_options.gff_outfile, StORF,
                                    StORF_Num)  # To keep consistency
                    if Reporter_options.annotation_type[1] == 'Out_Dir':
                        FASTA_StORF_write(Reporter_options, track_contig, fasta_outfile, StORF)
                    StORF_Num += 1
            Reporter_options.gff_outfile.write(line.strip() + '\n')

        elif line.startswith('##FASTA'):
            Reporter_options.gff_outfile.write(line.strip() + '\n')
            end = True
            # StORFs = find_after_StORFs(StORF_options, Contig_URS, track_prev_start, track_prev_stop, track_prev_contig)  # Changed to prev stop because we are switching from previous contig
            # if StORFs:
            #     for StORF in StORFs:
            #         GFF_StoRF_write(StORF_options, Reporter_options, track_prev_contig, outfile, StORF)  # To keep consistency
            #         if StORF_options.path == True:
            #             FASTA_StORF_write(Reporter_options, track_contig, fasta_outfile, StORF)
            #         StORF_Num += 1
            # if line != written_line:
            #     outfile.write(line.strip()+'\n')
            #     written_line = line

        else: # Now we just print out the remaining lines in the GFF
            Reporter_options.gff_outfile.write(line.strip()+'\n')
            written_line = line
##############################################################



def main():
    print("Thank you for using StORF-Reporter -- A detailed user manual can be found at https://github.com/NickJD/StORF-Reporter\nPlease report any issues to: https://github.com/NickJD/StORF-Reporter/issues\n#####")

    parser = argparse.ArgumentParser(description='StORF-Reporter ' + StORF_Reporter_Version + ': StORF-Reporter Run Parameters.', formatter_class=SmartFormatter)
    parser._action_groups.pop()

    required = parser.add_argument_group('Required Options')
    required.add_argument('-anno', action='store', dest='annotation_type', required=True,
                        choices=['Prokka', 'Bakta', 'Out_Dir', 'Single_GFF', 'Multiple_GFFs', 'Ensembl', 'Feature_Types',
                                  'Single_Genome', 'Multiple_Genomes','Single_Combined_GFF', 'Multiple_Combined_GFFs',
                                 'Pyrodigal', 'Single_FASTA', 'Multiple_FASTA'], nargs='*',
                        help='R|Select Annotation and Input options for one of the 3 options listed below\n'
                             '### Prokka/Bakta Annotation Option 1: \n'
                             '\tProkka = Report StORFs for a Prokka annotation; \n'
                             '\tBakta = Report StORFs for a Bakta annotation; \n'
                             '--- Prokka/Bakta Input Options: \n'
                             '\tOut_Dir = To provide the output directory of either a Prokka or Bakta run (will produce a new GFF and FASTA file); \n'
                             '\tSingle_GFF = To provide a single Prokka or Bakta GFF file (will not provide new FASTA file); \n'
                             '\tMultiple_GFFs = To provide a directory containing multiple GFF files in Prokka/Bakta format (will not provide a new FASTA file); \n\n'
                             
                             '### Standard GFF Annotation Option 2: \n'
                             '\tEnsembl = Report StORFs for an Ensembl Bacteria annotation (ID=gene); \n'
                             '\tFeature_Types = Used in conjunction with -gene_ident to define features such as CDS,rRNA,tRNA for UR extraction (default CDS); \n'
                             '--- Standard GFF Input Options: \n'
                             '\tSingle_Genome = To provide a single Genome - accompanying FASTA must share same name as given gff file (can be .fna, fa or .fasta); \n'
                             '\tMultiple_Genomes = To provide a directory containing multiple accompanying GFF and FASTA files - files must share the same name (fasta can be .fna, fa or .fasta); \n'
                             '\tSingle_Combined_GFF = To provide a GFF file with embedded FASTA at the bottom; \n'
                             '\tMultiple_Combined_GFFs = To provide a directory containing multiple GFF files with embedded FASTA at the bottom; \n\n'
                             
                             '### Complete Annotation Option 3: \n'
                             '\tPyrodigal = Run Pyrodigal then Report StORFs (provide path to single FASTA or directory of multiple FASTA files ;\n'
                             '--- Complete Annotation Input Options: \n'
                             '\tSingle_FASTA = To provide a single FASTA file; \n'
                             '\tMultiple_FASTA = To provide a directory containing multiple FASTA files (will detect .fna,.fa,.fasta); \n\n')

    required.add_argument('-p', action='store', dest='path', default='', required=True,
                        help='Provide input file or directory path')
    ###
    StORF_Reporter_args = parser.add_argument_group('StORF-Reporter Options')
    ### Add options to redirect into new output directory and filename - and output StORFs on their own in their own GFF -
    StORF_Reporter_args.add_argument('-oname', action="store", dest='o_name', required=False,
                        help='Default - Appends \'_StORF-R\' to end of input FASTA filename - Multiple_* runs will be numbered')
    StORF_Reporter_args.add_argument('-odir', action="store", dest='o_dir', required=False,
                        help='Default -  Same directory as input FASTA')
    StORF_Reporter_args.add_argument('-sout', action="store", dest='storfs_out', default=False, type=eval, choices=[True, False],
                        help='Default - False: Print out StORF sequences separately from Prokka/Bakta annotations')
    StORF_Reporter_args.add_argument('-lw', action="store", dest='line_wrap', default=True, type=eval, choices=[True, False],
                        help='Default - True: Line wrap FASTA sequence output at 60 chars')
    StORF_Reporter_args.add_argument('-aa', action="store", dest='translate', default=False, type=eval, choices=[True, False],
                        help='Default - False: Report StORFs as amino acid sequences')
    # output.add_argument('-gff_fasta', action="store", dest='gff_fasta', default=False, type=eval, choices=[True, False],
    #                     help='Default - False: Report all gene sequences (nt) at the bottom of GFF files in Prokka output mode')
    StORF_Reporter_args.add_argument('-gz', action='store', dest='gz', default=False, type=eval, choices=[True, False],
                        help='Default - False: Output as .gz')
    ###
    prodigal_options = parser.add_argument_group('Pyrodigal Options')
    prodigal_options.add_argument('-py_train', action='store', dest='py_train', required=False,
                        choices=['longest', 'individual', 'meta'], nargs='?', default='longest',
                        help='Default - longest: Type of model training to be done for Pyrodigal CDS prediction: '
                             'Options: longest = Trains on longest contig;'
                             ' individual = Trains on each contig separately - runs in meta mode if contig is < 20KB;'
                             ' meta = Runs in meta mode for all sequences')
    prodigal_options.add_argument('-py_fasta', action="store", dest='py_fasta', default=True, type=eval, choices=[True, False],
                                  help='Default - False: Output Pyrodigal+StORF predictions in FASTA format')
    prodigal_options.add_argument('-py_unstorfed', action="store", dest='py_unstorfed', default=False, type=eval, choices=[True, False],
                                  help='Default - False: Provide GFF containing original Pyrodigal predictions')

    ###
    UR_Ex_args = parser.add_argument_group('UR-Extractor Options')
    UR_Ex_args.add_argument('-gene_ident', action='store', dest='gene_ident', default='CDS',
                          help='Identifier used for extraction of Unannotated Regions "CDS,rRNA,tRNA" - To be used with "-anno Feature_Types"')
    UR_Ex_args.add_argument('-min_len', action='store', dest='minlen', default='30', type=int,
                          help='Default - 30: Minimum UR Length')
    UR_Ex_args.add_argument('-max_len', action='store', dest='maxlen', default='100000', type=int,
                          help='Default - 100,000: Maximum UR Length')
    UR_Ex_args.add_argument('-ex_len', action='store', dest='exlen', default='50', type=int,
                          help='Default - 50: UR Extension Length')
    ###
    StORF_Finder_args = parser.add_argument_group('StORF-Finder Options')
    StORF_Finder_args.add_argument('-spos', action="store", dest='stop_inclusive', default=False, type=eval,
                          choices=[True, False],
                          help='Default - False: Output StORF positions inclusive of first stop codon')
    StORF_Finder_args.add_argument('-rs', action="store", dest='remove_stop', default=True, type=eval, choices=[True, False],
                        help='Default - True: Remove stop "*" from StORF amino acid sequences')

    StORF_Finder_args.add_argument('-con_storfs', action="store", dest='con_storfs', default=False, type=eval, choices=[True, False],
                        help='Default - False: Output Consecutive StORFs')
    StORF_Finder_args.add_argument('-con_only', action="store", dest='con_only', default=False, type=eval, choices=[True, False],
                        help='Default - False: Only output Consecutive StORFs')
    StORF_Finder_args.add_argument('-ps', action="store", dest='partial_storf', default=False, type=eval, choices=[True, False],
                        help='Default - False: Partial StORFs reported')
    StORF_Finder_args.add_argument('-wc', action="store", dest='whole_contig', default=False, type=eval, choices=[True, False],
                        help='Default - False: StORFs reported across entire sequence')
    StORF_Finder_args.add_argument('-short_storfs', action="store", dest='short_storfs', default=False, type=str, choices=['False', 'Nolap', 'Olap'],
                        help='Default - False: Run StORF-Finder in "Short-StORF" mode. Will only return StORFs between 30 and 120 nt '
                             'that do not overlap longer StORFs - Only works with StORFs for now. "Nolap" will filter Short-StORFs which are'
                             'overlapped by StORFs and Olap will report Short-StORFs which do overlap StORFs. Overlap is defined by "-olap".')
    StORF_Finder_args.add_argument('-short_storfs_only', action="store", dest='short_storfs_only', default=False, type=eval, choices=[True, False],
                        help='Default - True. Only report Short-StORFs?')
    StORF_Finder_args.add_argument('-minorf', action="store", dest='min_orf', default=99, type=int,
                          help='Default - 99: Minimum StORF size in nt')
    StORF_Finder_args.add_argument('-maxorf', action="store", dest='max_orf', default=60000, type=int,
                        help='Default - 60kb: Maximum StORF size in nt')
    StORF_Finder_args.add_argument('-codons', action="store", dest='stop_codons', default="TAG,TGA,TAA",
                          help='Default - (\'TAG,TGA,TAA\'): List Stop Codons to use')
    StORF_Finder_args.add_argument('-olap_filt', action='store', dest='olap_filtering', default='both-strand',
                          const='both-strand', nargs='?',
                          choices=['none', 'single-strand', 'both-strand'],
                          help='Default - "both-strand": Filtering level "none" is not recommended, "single-strand" for single strand filtering '
                               'and both-strand for both-strand longest-first tiling')
    StORF_Finder_args.add_argument('-start_filt', action="store", dest='start_filtering', default=False, type=eval,
                          choices=[True, False],
                          help='Default - False: Filter out StORFs without at least one of the 3 common start codons (best used for short-storfs).')
    StORF_Finder_args.add_argument('-so', action="store", dest='storf_order', default='start_pos', nargs='?', choices=['start_pos','strand'],
                        required=False,
                        help='Default - Start Position: How should StORFs be ordered when >1 reported in a single UR.')
    StORF_Finder_args.add_argument('-f_type', action='store', dest='feature_type', default='CDS', const='CDS', nargs='?',
                          choices=['StORF', 'CDS', 'ORF'],
                          help='Default - "CDS": Which GFF feature type for StORFs to be reported as in GFF - '
                               '"CDS" is probably needed for use in tools such as Roary and Panaroo')
    StORF_Finder_args.add_argument('-olap', action="store", dest='overlap_nt', default=50, type=int,
                          help='Default - 50: Maximum number of nt of a StORF which can overlap another StORF.')
    StORF_Finder_args.add_argument('-ao', action="store", dest='allowed_overlap', default=50, type=int,
                          help='Default - 50 nt: Maximum overlap between a StORF and an original gene.')
    # optional.add_argument('-col', action='store', dest='genome_collection', default='', required=False,
    #                     help='Provide directories containing a collection of matching FASTA and GFF files'
    #                          '(list .fa then .gff containing directories separated by commas - ./FA,./GFF)')
    #optional.add_argument('-comb', action='store', dest='combined_gffs', default='', required=False,
    #                    help='Provide directory containing GFFs with sequences combined into single file to be StORFed - Only produces modified GFFs')

    misc = parser.add_argument_group('Misc')


    misc.add_argument('-overwrite', action='store', dest='overwrite', default=False, type=eval, choices=[True, False],
                        help='Default - False: Overwrite StORF-Reporter output if already present')
    misc.add_argument('-v', action='store', dest='verbose', default=False, type=eval, choices=[True, False],
                        help='Default - False: Print out runtime messages')
    parser.add_argument('-nout', action='store', dest='nout', default=True, type=eval, choices=[True, False],
                        help=argparse.SUPPRESS)
    parser.add_argument('-nout_pyrodigal', action='store', dest='nout_pyrodigal', default=True, type=eval, choices=[True, False],
                        help=argparse.SUPPRESS)

    Reporter_options = parser.parse_args()
    Reporter_options.gff_fasta = None # To be done
    Reporter_options.reporter = True

    ##############
    ## Print out user chosen annotation options

    ##############
    ## Incompatible argument catching
    if len(Reporter_options.annotation_type) != 2:
        parser.error('Please select two compatible options for required argument -anno.')
    ##############
    #### Output Directory and Filename handling
    ### Add '/' to end of directory path
    if Reporter_options.annotation_type[1] in ['Multiple_GFFs','Multiple_Genomes','Multiple_FASTA','Out_Dir','Comb_GFFs', 'Multiple_Combined_GFFs']:
        Reporter_options.path = Reporter_options.path + '/' if not Reporter_options.path.endswith('/') else Reporter_options.path

    if Reporter_options.o_dir == None and Reporter_options.o_name == None:
        if Reporter_options.annotation_type[1] == 'Out_Dir':
            output_file = Reporter_options.path.split('.')[-1]
            output_file = output_file + output_file.split('/')[-2] + '_StORF-Reporter_Extended'
        else:
            tmp_extension = Reporter_options.path.split('.')[-1]  # could be .fa/.fasta etc
            output_file = Reporter_options.path.replace('.' + tmp_extension, '')
            if Reporter_options.annotation_type[0] == 'Pyrodigal':
                output_file = output_file + '_Pyrodigal_StORF-Reporter_Extended'
            else:
                output_file = output_file + '_StORF-Reporter_Extended'
    elif Reporter_options.o_dir != None and Reporter_options.o_name != None:
        output_file = Reporter_options.o_dir
        output_file = output_file + '/' if not output_file.endswith('/') else output_file
        output_file = output_file + Reporter_options.o_name
    elif Reporter_options.o_dir != None:
        tmp_extension = Reporter_options.path.split('.')[-1]  # could be .fa/.fasta etc
        output_file = Reporter_options.path.replace('.' + tmp_extension, '').split('/')[-1]
        if Reporter_options.annotation_type[0] == 'Pyrodigal':
            output_file = Reporter_options.o_dir + output_file + '_Pyrodigal_StORF-Reporter_Extended'
        else:
            output_file = Reporter_options.o_dir + output_file + '_StORF-Reporter_Extended'
    elif Reporter_options.o_name != None:
        tmp_filename = Reporter_options.path.split('/')[-1]  # could be .fa/.fasta etc
        output_file = Reporter_options.path.replace(tmp_filename, '')
        output_file = output_file + Reporter_options.o_name

###############################################################################################

    ############## Setup for Prokka/Bakta output directory
    if Reporter_options.annotation_type[0] in ('Prokka','Bakta') and Reporter_options.annotation_type[1] == 'Out_Dir':
        Reporter_options.output_file = output_file
        #### Checking and cleaning
        for fname in os.listdir(Reporter_options.path):
            if '_StORF-Reporter_Extended' in fname and Reporter_options.overwrite == False:
                parser.error(
                    'Prokka/Bakta directory not clean and already contains a StORF-Reporter output. Please delete or use "-overwrite True" and try again.')
            elif '_StORF-Reporter_Extended' in fname and Reporter_options.overwrite == True:
                os.remove(Reporter_options.path + '/' + fname)
                if Reporter_options.verbose == True:
                    print('StORF-Reporter output ' + fname + ' will be overwritten.')
        ####
        Reporter_options.gene_ident = "misc_RNA,gene,mRNA,CDS,rRNA,tRNA,tmRNA,CRISPR,ncRNA,regulatory_region,oriC,pseudo"
        Contigs, Reporter_options = run_UR_Extractor_Directory(Reporter_options)
        ################## Find StORFs in URs - Setup StORF_Reporter-Finder Run
        Reporter_StORFs = StORF_Reported(Reporter_options, Contigs)
        StORF_Filler(Reporter_options, Reporter_StORFs)
        if Reporter_options.verbose == True:
            print("Finished: " + Reporter_options.gff.split('.')[0])

    ############### Setup for directory of a single Prokka/Bakta GFF or a directory of multiple Prokka/Bakta GFFs
    elif Reporter_options.annotation_type[0] in ('Prokka', 'Bakta') and Reporter_options.annotation_type[1]  in ('Single_GFF', 'Multiple_GFFs'):
        if Reporter_options.annotation_type[1]  == "Single_GFF":
            gff_list = [Reporter_options.path]
        elif Reporter_options.annotation_type[1]  == "Multiple_GFFs":
            gff_list = list(pathlib.Path(Reporter_options.path).glob('*.gff'))
            gff_list.extend(pathlib.Path(Reporter_options.path).glob('*.gff3'))
        #### Checking and cleaning
        gff_list = list(map(str, gff_list))
        gffs_to_filter = []
        for gff in gff_list:
            if '_StORF-Reporter_Extended.gff' in gff and os.path.isfile(gff) and Reporter_options.overwrite == False:
                gffs_to_filter.append(gff)
                gffs_to_filter.append(gff.replace('_StORF-Reporter_Extended.gff','.gff'))
                print('Prokka/Bakta GFF has already been processed and a StORF-Reporter output exists for ' + gff.split('/')[-1] + '. Please delete or use "-overwrite True" and try again.')
            elif '_StORF-Reporter_Extended.gff' in gff and os.path.isfile(gff) and Reporter_options.overwrite == True:
                os.remove(gff)
                gffs_to_filter.append(gff)
                if Reporter_options.verbose == True:
                    print('StORF-Reporter output '  + gff.split('/')[-1] +  ' will be overwritten.')
        gff_list = [x for x in gff_list if x not in gffs_to_filter]
        ####
        Reporter_options.gene_ident = "misc_RNA,gene,mRNA,CDS,rRNA,tRNA,tmRNA,CRISPR,ncRNA,regulatory_region,oriC,pseudo"
        file_counter = 0
        for gff in gff_list:
            # Finalising output_file name
            if Reporter_options.annotation_type[1]  == "Multiple_GFFs" and Reporter_options.o_name != None:
                Reporter_options.output_file = output_file + '_' + str(file_counter)
                file_counter += 1
            elif Reporter_options.annotation_type[1]  == "Multiple_GFFs":
                tmp_filename = gff.split('/')[-1].split('.gff')[0]  # could be .fa/.fasta etc
                Reporter_options.output_file = output_file.replace('_StORF-Reporter',tmp_filename + '_StORF-Reporter')
            else:
                Reporter_options.output_file = output_file
            if Reporter_options.verbose == True:
                print("Starting: " + str(gff))
            Contigs, Reporter_options = run_UR_Extractor_Extended_GFFs(Reporter_options,gff)
            ################## Find StORFs in URs - Setup StORF_Reporter-Finder Run
            Reporter_StORFs = StORF_Reported(Reporter_options, Contigs)
            StORF_Filler(Reporter_options, Reporter_StORFs)
            if Reporter_options.verbose == True:
                print("Finished: " + gff.split('/')[-1]) # Will add number of additional StORFs here

    ################ Run StORF-Reporter on standardGFFs
    elif Reporter_options.annotation_type[0] in ('Ensembl', 'Feature_Types'):
        if Reporter_options.annotation_type[0] == 'Ensembl':
            Reporter_options.gene_ident = "ID=gene"
        elif Reporter_options.annotation_type[0] == 'Feature_Types':
            Reporter_options.gene_ident = Reporter_options.gene_ident.split(',')
        ####
        if Reporter_options.annotation_type[1] in ("Single_Genome", "Single_Combined_GFF"):
            gff_list = [Reporter_options.path]
        elif Reporter_options.annotation_type[1] in ("Multiple_Genomes", "Multiple_Combined_GFFs"):
            gff_list = list(pathlib.Path(Reporter_options.path).glob('*.gff'))
            gff_list.extend(pathlib.Path(Reporter_options.path).glob('*.gff3'))
        else:
            parser.error('Please select two compatible options for required argument -anno.')
        #### Checking and cleaning
        gff_list = list(map(str, gff_list))
        gffs_to_filter = []
        for gff in gff_list:
            if '_StORF-Reporter_Extended.gff' in gff and os.path.isfile(gff) and Reporter_options.overwrite == False:
                gffs_to_filter.append(gff)
                gffs_to_filter.append(gff.replace('_StORF-Reporter_Extended.gff', '.gff'))
                print('Prokka/Bakta GFF has already been processed and a StORF-Reporter output exists for ' +
                      gff.split('/')[-1] + '. Please delete or use "-overwrite True" and try again.')
            elif '_StORF-Reporter_Extended.gff' in gff and os.path.isfile(gff) and Reporter_options.overwrite == True:
                os.remove(gff)
                gffs_to_filter.append(gff)
                if Reporter_options.verbose == True:
                    print('StORF-Reporter output ' + gff.split('/')[-1] + ' will be overwritten.')
        gff_list = [x for x in gff_list if x not in gffs_to_filter]
        ####
        file_counter = 0
        for gff in gff_list:
            # Finalising output_file name
            if Reporter_options.annotation_type[1]  == "Multiple_Combined_GFFs" and Reporter_options.o_name != None:
                Reporter_options.output_file = output_file + '_' + str(file_counter)
                file_counter += 1
            elif Reporter_options.annotation_type[1]  == "Multiple_Combined_GFFs":
                tmp_filename = gff.split('/')[-1].split('.gff')[0]  # could be .fa/.fasta etc
                Reporter_options.output_file = output_file.replace('_StORF-Reporter',tmp_filename + '_StORF-Reporter')
            else:
                Reporter_options.output_file = output_file


            if Reporter_options.verbose == True:
                print("Starting: " + str(gff))
            if Reporter_options.annotation_type[1] in ('Single_Combined_GFF', 'Multiple_Combined_GFFs'):
                Contigs, Reporter_options = run_UR_Extractor_Extended_GFFs(Reporter_options, gff)
            else:
                Contigs, Reporter_options = run_UR_Extractor_Matched(Reporter_options, gff)
            ################## Find StORFs in URs - Setup StORF_Reporter-Finder Run
            Reporter_StORFs = StORF_Reported(Reporter_options, Contigs)
            StORF_Filler(Reporter_options, Reporter_StORFs)
            if Reporter_options.verbose == True:
                print("Finished: " + gff.split('/')[-1])  # Will add number of additional StORFs here`

    ####### Run Pyrodigal and then StORF-Reporter on either a single or directory of FASTA files
    elif Reporter_options.annotation_type[0] == 'Pyrodigal' and Reporter_options.annotation_type[1] in ("Single_FASTA", "Multiple_FASTA"):  # needs cleaning
        Reporter_options.gene_ident = "gene"
        if os.path.isdir(Reporter_options.path) and Reporter_options.annotation_type[1]  == "Multiple_FASTA":
            fasta_list = list(pathlib.Path(Reporter_options.path).glob('*.fasta'))
            fasta_list.extend(list(pathlib.Path(Reporter_options.path).glob('*.fna')))
            fasta_list.extend(list(pathlib.Path(Reporter_options.path).glob('*.fa')))
            fasta_list = list(map(str, fasta_list))
        elif os.path.isfile(Reporter_options.path) and Reporter_options.annotation_type[1]  == "Single_FASTA":
            fasta_list = [Reporter_options.path]
        else:
            parser.error('Please provide a directory or file path to a genomic FASTA file.')
        file_counter = 0
        for fasta in fasta_list:
            # Finalising output_file name
            if Reporter_options.annotation_type[1]  == "Multiple_FASTA" and Reporter_options.o_name != None:
                Reporter_options.output_file = output_file + '_' + str(file_counter)
                file_counter += 1
            elif Reporter_options.annotation_type[1]  == "Multiple_FASTA":
                Reporter_options.output_file = output_file.replace('_Pyrodigal',fasta.name.split('.')[0] + '_Pyrodigal')
            else:
                Reporter_options.output_file = output_file
            if Reporter_options.py_unstorfed == True:  # User wants to output initial Pyrodigal GFF
                Reporter_options.pyrodigal_out_file_fasta = open(Reporter_options.output_file + '_Pyrodigal_Predictions.fasta','w')
                Reporter_options.pyrodigal_out_file_gff = open(Reporter_options.output_file + '_Pyrodigal_Predictions.gff', 'w')
            if Reporter_options.verbose == True:
                print("Starting: " + str(fasta))
            ###### Run Pyrodigal
            Contigs, Reporter_options = pyrodigal_predict(str(fasta),Reporter_options)
            ###### Find StORFs in URs - Setup StORF_Reporter-Finder Run
            Reporter_StORFs = StORF_Reported(Reporter_options, Contigs)
            ###### Append StORFs to current GFF file from provided genome annotation
            StORF_Filler(Reporter_options, Reporter_StORFs)
            print("Finished predicting Pyrodigal CDS genes and StORFs for -  " + fasta.split('/')[-1]) # Can add number of additional StORFs here
    else:
        parser.error('Incompatible user-arguments selected.')




if __name__ == "__main__":
    main()
    print("Complete")





import argparse
import collections
from datetime import date
import gzip
import sys
import os
import pathlib


try:
    from .Constants import *
except (ModuleNotFoundError, ImportError, NameError, TypeError) as error:
    from Constants import *

###################
gencode = {
      'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M',
      'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T',
      'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K',
      'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R',
      'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L',
      'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P',
      'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q',
      'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R',
      'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V',
      'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A',
      'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E',
      'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G',
      'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S',
      'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L',
      'TAC':'Y', 'TAT':'Y', 'TAA':'*', 'TAG':'*',
      'TGC':'C', 'TGT':'C', 'TGA':'*', 'TGG':'W'}

def translate_frame(sequence):
    translate = ''.join([gencode.get(sequence[3 * i:3 * i + 3], 'X') for i in range(len(sequence) // 3)])
    return translate


def reverseCorrectLoci(seq_length,first,second,third): # here for the negative loci correction

    if second == None:
        corrected_start = max(seq_length - int(third),1)
        corrected_stop = max(seq_length - int(first-1),1)
        return corrected_start, corrected_stop
    else: # Needs to be checked
        corrected_start = max(seq_length - int(third),1)
        corrected_mid = max(seq_length - int(second-3),1)
        corrected_stop = max(seq_length - int(first),1)
        return corrected_start, corrected_mid, corrected_stop
#################################
def revCompIterative(watson): #Gets Reverse Complement

    complements = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 'N': 'N',
                   'R': 'Y', 'Y': 'R', 'S': 'S', 'W': 'W', 'K': 'M',
                   'M': 'K', 'V': 'B', 'B': 'V', 'H': 'D', 'D': 'H'}
    watson = watson.upper()
    watsonrev = watson[::-1]
    crick = ""
    for nt in watsonrev: # make dict to catch bad nts - if more than 1 output them to std error
        try:
            crick += complements[nt]
        except KeyError:
            crick += nt
            #ns_nt[nt] +=1
    return crick


def write_fasta(dna_regions, fasta_outfile):
    for dna_region, dna_region_ur in dna_regions.items():
        fasta_outfile.write('##sequence-region\t' + dna_region + ' 1 ' + str(len(dna_region_ur[0])) + '\n')
        if dna_region_ur[3]:
            for storf, seq in dna_region_ur[3].items():
                fasta_outfile.write('>Start=' + storf.split('_')[0] + ';Stop=' + storf.split('_')[1] + ';Frame=' +
                                storf.split('_')[2] + storf.split(';')[1] + '\n' + seq + '\n')
    fasta_outfile.close()

def write_gff(dna_regions,options,gff_outfile, gff):
    gff_outfile.write("##gff-version\t3\n#\tStORF-Extractor \n#\tRun Date:" + str(date.today()) + '\n')
    gff_outfile.write('##StORF-Reporter ' + StORF_Reporter_Version + '\n')
    for seq_reg in dna_regions:
        gff_outfile.write('##sequence-region\t' + seq_reg + ' 1 ' + str(dna_regions[seq_reg][1]) + '\n')
    gff_outfile.write("##Original File: " + gff.split('/')[-1] + '\n\n')
    for dna_region, storf_data in dna_regions.items():
        #ur_ident = dna_region + options.ident
        if storf_data[3]:
            for storf, seq in storf_data[3].items():

                entry = (dna_region + '\tStORF-Reporter\tCDS\t' + storf.split('_')[0] + '\t' + storf.split('_')[1]+ '\t.\t' +
                         storf.split('_')[2].split(';')[0]  + '\t.\t' + storf.split(';',1)[1] + '\n')
                gff_outfile.write(entry)
    gff_outfile.close()

def fasta_load(fasta_in):
    dna_regions = collections.OrderedDict()
    first = True
    if '>' in fasta_in.readline().rstrip():
        fasta_in.seek(0)
        #### Default for when presented with standard fasta file
        for line in fasta_in:
            line = line.strip()
            if line.startswith('>') and first == False:  # Check if first seq in file
                dna_region_length = len(seq)
                dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
                seq = ''
                dna_region_id = line.split()[0].replace('>', '')
            elif line.startswith('>'):
                seq = ''
                dna_region_id = line.split()[0].replace('>', '')
            else:
                seq += str(line)
                first = False
        dna_region_length = len(seq)
        dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
    elif '##' in  fasta_in.readline().rstrip(): # Clunky and may fall over
        fasta_in.seek(0)
        #### Called when presented with Prokka GFF file so must get fasta from inside it
        ### Get to genome seq
        at_FASTA = False
        for line in fasta_in:  # Get gene loci from GFF - ID=Gene will also classify Pseudogenes as genes
            if line.startswith('##FASTA'):  # Not to crash on empty lines in GFF
                at_FASTA = True
            elif at_FASTA == True:
                line = line.strip()
                if line.startswith('>') and first == False:  # Check if first seq in file
                    dna_region_length = len(seq)
                    dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
                    seq = ''
                    dna_region_id = line.split()[0].replace('>', '')
                elif line.startswith('>'):
                    seq = ''
                    dna_region_id = line.split()[0].replace('>', '')
                else:
                    seq += str(line)
                    first = False
        dna_region_length = len(seq)
        dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})

    return dna_regions


def gff_load(options,gff_in,dna_regions):
    has_storfs = False
    for line in gff_in:  # Get gene loci from GFF - ID=Gene will also classify Pseudogenes as genes
        line_data = line.split()
        if line.startswith('\n') or line.startswith('#'):  # Not to crash on empty lines in GFF
            continue
        else:
            try:
                if line_data[0] in dna_regions:
                    if any(tool in line_data[1] for tool in options.tool_ident):
                        has_storfs = True
                        pos = line_data[3] + '_' + line_data[4] + '_' + line_data[6] + ';' + line_data[8]
                        if pos not in dna_regions[line_data[0]][2]:
                            dna_regions[line_data[0]][2].append(pos) # This will add to list
            except IndexError:
                continue
    if has_storfs == False:
        dna_regions = None
    return dna_regions


def run_StORF_Extractor_Combined_GFFs(options,gff): # When given a directory with multiple GFFs but without accompianing .fna
    gff = str(gff)
    options.gff = gff
    options.fasta = gff
    storf_extractor(options, gff)
    #return StORFs, options

def run_StORF_Extractor_Matched(options,gff): # When given a directory with multiple GFFs but with accompianing .fna
    gff = str(gff)
    options.gff = gff
    fasta = gff.replace('.gff','.fasta')
    fna = gff.replace('.gff', '.fna')
    fa = gff.replace('.gff', '.fa')
    if os.path.isfile(fasta):
        options.fasta = fasta
    elif os.path.isfile(fna):
        options.fasta = fna
    elif os.path.isfile(fa):
        options.fasta = fa
    else:
        sys.exit('No matching FASTA file found for ' + gff)
    storf_extractor(options, gff)
    #return StORFs, options


def storf_extractor(options, gff):
    try:
        try: # Detect whether fasta/gff files are .gz or text and read accordingly
            fasta_in = gzip.open(options.fasta,'rt')
            dna_regions = fasta_load(fasta_in)
        except:
            fasta_in = open(options.fasta,'r',encoding='unicode_escape')
            dna_regions = fasta_load(fasta_in)
        try:
            gff_in = gzip.open(options.gff,'rt')
            dna_regions = gff_load(options,gff_in,dna_regions)
        except:
            gff_in = open(options.gff,'r',encoding='unicode_escape')
            dna_regions = gff_load(options,gff_in,dna_regions)
    except AttributeError:
        sys.exit("Attribute Error:\nStORF'ed GFF probably already exists - Must be deleted before running")
    if dna_regions == None and options.verbose == True:
        print("No StORFs to extract from " + gff)
        os.remove(options.fasta_outfile.name)
        os.remove(options.gff_outfile.name)
        return
    elif dna_regions == None:
        os.remove(options.fasta_outfile.name)
        if options.gff_out == True:
            os.remove(options.gff_outfile.name)
        return


    for (key,(seq,seq_length,posns,tmp))  in dna_regions.items(): #Extract URs from 1 dna_region at a time
        Extracted_StORFs = collections.OrderedDict()
        seq_rev = revCompIterative(seq)
        if posns: # If UR has a pos
            for pos in posns: # Iterate over GFF loci and measure flanking regions for potential URs
                start = int(pos.split('_')[0])
                stop = int(pos.split('_')[1])
                frame = pos.split('_')[2].split(';')[0]
                ###### This hack is to get over GFF errors where genome-long annotations
                if stop-start >= 100000:
                    if options.verbose == True:
                        print("UR " + pos + " is more than 100,000 kbs - Please Check Annotation")
                    continue
                if frame == '+':
                    StORF_seq = seq[start:stop]
                elif frame == '-':
                    rev_corrected_start, rev_corrected_stop = reverseCorrectLoci(seq_length, start, None, stop)
                    StORF_seq = seq_rev[rev_corrected_start:rev_corrected_stop]

                Extracted_StORFs[pos] = StORF_seq
        dna_regions.update({key: (seq, seq_length, posns, Extracted_StORFs)})



    write_fasta(dna_regions, options.fasta_outfile)
    if options.gff_out != False:
        write_gff(dna_regions, options, options.gff_outfile, gff)

def main():
    print("Thank you for using StORF-Reporter -- A detailed user manual can be found at https://github.com/NickJD/StORF-Reporter\nPlease report any issues to: https://github.com/NickJD/StORF-Reporter/issues\n#####")

    parser = argparse.ArgumentParser(description='StORF-Reporter ' + StORF_Reporter_Version + ': StORF-Extractor Run Parameters.')
    parser._action_groups.pop()

    required = parser.add_argument_group('Required Arguments')
    required.add_argument('-storf_input', action='store', dest='storf_input', required=True,
                        choices=['Combined', 'Separate'],
                        help='Are StORFs to be extracted from Combined GFF/FASTA or Separate GFF/FASTA files?\n')
    required.add_argument('-p', action='store', dest='path', default='', required=True,
                        help='Provide input file or directory path')

    ### Not implemented yet
    # optional = parser.add_argument_group('Optional Arguments')
    # optional.add_argument('-tool_ident', action='store', dest='tool_ident', default='StORF-Reporter',
    #                         help='Default "StORF-Reporter": Tool-name Identifier used for extraction of StORFs or other genomic elements'
    #                              ' "StORF-Reporter, Prodigal, Pyrodigal" (second GFF column)')


    output = parser.add_argument_group('Output')
    output.add_argument('-gff_out', action='store', dest='gff_out', default=False, type=eval, choices=[True, False],
                        help='Default - False: Output StORFs in GFF format')
    output.add_argument('-oname', action="store", dest='o_name', required=False,
                        help='Default - Appends \'_Extracted_StORFs\' to end of input GFF filename')
    output.add_argument('-odir', action="store", dest='o_dir', required=False,
                        help='Default -  Same directory as input FASTA')
    output.add_argument('-gz', action='store', dest='gz', default='False', type=eval, choices=[True, False],
                        help='Default - False: Output as .gz')


    misc = parser.add_argument_group('Misc')
    misc.add_argument('-v', action='store', dest='verbose', default='False', type=eval, choices=[True, False],
                        help='Default - False: Print out runtime status')



    options = parser.parse_args()
    options.tool_ident = ['StORF-Reporter']#options.tool_ident.split(',')

    #### Output Directory and Filename handling
    if options.o_dir == None and options.o_name == None:
        tmp_extension = options.path.split('.')[-1]  # could be .fa/.fasta etc
        output_file = options.path.replace('.' + tmp_extension, '')
        output_file = output_file + '_Extracted_StORFs'
    elif options.o_dir != None and options.o_name != None:
        output_file = options.o_dir
        output_file = output_file + '/' if not output_file.endswith('/') else output_file
        output_file = output_file + options.o_name
    elif options.o_dir != None:
        tmp_extension = options.path.split('.')[-1]  # could be .fa/.fasta etc
        output_file = options.path.replace('.' + tmp_extension, '').split('/')[-1]
        output_file = options.o_dir + output_file + '_Extracted_StORFs'
    elif options.o_name != None:
        tmp_filename = options.path.split('/')[-1]  # could be .fa/.fasta etc
        output_file = options.path.replace(tmp_filename, '')
        output_file = output_file + options.o_name

######################################################################
    if os.path.isdir(options.path):
        options.path = options.path + '/' if not options.path.endswith('/') else options.path
        gff_list = list(pathlib.Path(options.path).glob('*.gff'))
        gff_list.extend(pathlib.Path(options.path).glob('*.gff3'))
    else:
        gff_list = [options.path]
    gff_list = list(map(str, gff_list))
    ####
    file_counter = 0
    for gff in gff_list:
        ### Checking if file has already been StORF-Extracted
        with open(gff) as f:
            second_line = f.readlines()[1]
            if 'StORF-Extractor' in second_line:
                if options.verbose == True:
                    print(gff + " is an extracted file and will be ignored.")
                continue
            #f.seek(0) # needed?
        # Finalising output_file name
        if os.path.isdir(options.path) and options.o_name != None:
            tmp_output_file = output_file + '_' + str(file_counter)
            file_counter += 1
        elif os.path.isdir(options.path):
            tmp_filename = gff.split('/')[-1].split('.gff')[0]  # could be .fa/.fasta etc
            tmp_output_file = output_file.replace('_Extracted_StORFs', tmp_filename + '_Extracted_StORFs')
        else:
            tmp_output_file = output_file

        if options.gz == False:
            options.fasta_outfile = open(tmp_output_file + '.fasta', 'w', newline='\n', encoding='utf-8')
            if options.gff_out == True:
                options.gff_outfile = open(tmp_output_file + '.gff', 'w', newline='\n', encoding='utf-8')
        elif options.gz == True:
            options.fasta_outfile = open(tmp_output_file + '.fasta.gz', 'wt', newline='\n', encoding='utf-8')
            if options.gff_out == True:
                options.gff_outfile = open(tmp_output_file + '.gff.gz', 'wt', newline='\n', encoding='utf-8')

        if options.verbose == True:
            print("Starting: " + str(gff))
        if  options.storf_input == 'Separate':
            run_StORF_Extractor_Matched(options, gff)
        else:
            run_StORF_Extractor_Combined_GFFs(options, gff)


        ##################

        if options.verbose == True:
            print("Finished: " + gff.split('/')[-1])  # Will add number of additional StORFs here`




if __name__ == "__main__":
    main()
    print("Complete")





























import argparse
import numpy as np

parser = argparse.ArgumentParser()
parser.add_argument('-ur', '--URs', required= True, help='Which URs to analyse?')
args = parser.parse_args()

def seq_Lengths(seqs):
    seq = ""
    lengths = []
    prev_ID = ""
    with open(seqs, 'r') as seqs:
        for line in seqs:
            line = line.replace("\n","")
            if line.startswith('>'):
                # if len(seq) >= 100000:
                #     print(prev_ID)
                if len(seq) != 0:
                    lengths.append(len(seq))
                    if len(seq) <=50:
                        print("")
                seq = ""
                prev_ID = line
            if not line.startswith('>'):
                seq += str(line)



    print("Number of Seqs: " + str(len(lengths)))
    print("Longest Seq: " + str(max(lengths)))
    print("Shortest Seq: " + str(min(lengths)))
    print("Median of Seqs: " + str(np.median(lengths)))
    print("Mean of Seqs: " + format(np.mean(lengths), '.2f'))
    print("Std: " + format(np.std(lengths),'.2f'))
    print("25th Percentile " + format(np.percentile(lengths, 25), '.2f'))
    print("75th Percentile " + format(np.percentile(lengths,75),'.2f'))


if __name__ == "__main__":
    seq_Lengths(**vars(args))



import argparse
import collections
import re
from collections import defaultdict, OrderedDict
from datetime import date
import textwrap
import gzip
import os
import sys

try:
    from ORForise.utils import sortORFs  # Calling from ORForise via pip
    from .Constants import *
except (ModuleNotFoundError, ImportError, NameError, TypeError) as error:
    sys.path.insert(0, '../../../ORForise/src/ORForise/') # Calling from ORForise locally (StORF_Reporter and ORForise in same dir)
    from utils import sortORFs
    from Constants import *




###################
gencode = {
      'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M',
      'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T',
      'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K',
      'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R',
      'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L',
      'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P',
      'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q',
      'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R',
      'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V',
      'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A',
      'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E',
      'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G',
      'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S',
      'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L',
      'TAC':'Y', 'TAT':'Y', 'TAA':'*', 'TAG':'*',
      'TGC':'C', 'TGT':'C', 'TGA':'*', 'TGG':'W'}

def translate_frame(sequence):
    translate = ''.join([gencode.get(sequence[3 * i:3 * i + 3], 'X') for i in range(len(sequence) // 3)])
    return translate

############################

def reverseCorrectLoci(sequence_id,first,second,third): # here for the negative loci correction
    ur_length = int(sequence_id.split('_')[-1]) - int(sequence_id.split('_')[-2])
    if second == None:
        corrected_start = max(ur_length - int(third-3),1)
        corrected_stop = max(ur_length - int(first-3),1)
        return corrected_start, corrected_stop
    else:
        corrected_start = max(ur_length - int(third-3),1)
        corrected_mid = max(ur_length - int(second-3),1)
        corrected_stop = max(ur_length - int(first-3),1)
        return corrected_start, corrected_mid, corrected_stop


#################################
def revCompIterative(watson): #Gets Reverse Complement

    complements = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 'N': 'N',
                   'R': 'Y', 'Y': 'R', 'S': 'S', 'W': 'W', 'K': 'M',
                   'M': 'K', 'V': 'B', 'B': 'V', 'H': 'D', 'D': 'H'}
    watson = watson.upper()
    watsonrev = watson[::-1]
    crick = ""
    for nt in watsonrev: # make dict to catch bad nts - if more than 1 output them to std error
        try:
            crick += complements[nt]
        except KeyError:
            crick += nt
            #ns_nt[nt] +=1
    return crick

######## Might only be the stop which is the first constorfs end.
def prev_con_StORF_CHECKER(prev_con_StORF,sequence,options):
    last_stop_str = int(prev_con_StORF.split(',')[-1])
    seq = sequence[last_stop_str:last_stop_str + 3]
    if not any(seq in s for s in options.stop_codons.split(',')):
        last_stop = int(last_stop_str) - 3
        last_stop = str(last_stop)
        prev_con_StORF = prev_con_StORF.replace(str(last_stop_str), last_stop)
    return prev_con_StORF

#########
def cut_seq(wc_seq,end):
    while len(wc_seq) % 3 != 0:
        if '+' in end:
            wc_seq = wc_seq[:-1] # keep removing char
        elif '-' in end:
            wc_seq = wc_seq[1:]  # keep removing char
    return wc_seq

def start_filtering(storfs):
    keep_storfs = {}
    for pos, data in storfs.items():
        starts = []
        for start_codon in ['ATG', 'GTG', 'TTG']:
            starts += [match.start() for match in re.finditer(re.escape(start_codon), data[0])]
        #### This must be slow?
        for start in starts:
            if start % 3 == 0: # maybe add a clause to report only those storfs with starts in first n of storfs
                keep_storfs.update({pos:data})
                break
    return keep_storfs

def tile_filtering(storfs,options): #both-strand filtering
    #storfs = OrderedDict(sorted(storfs.items(), key=lambda e: tuple(map(int, e[0].split(","))))) # Is this needed if I am reordering below?
    ################ - Order largest first filtering
    storfs = sorted(storfs.items(), key=lambda storfs:storfs[1][3],reverse=True)
    ordered_by_length = OrderedDict()
    for tup in storfs:
        ordered_by_length.update({tup[0]:tup[1]}) # -1 for last?
    ############## - For each StORF_Reporter, remove all smaller overlapping STORFs according to filtering rules
    num_storfs = len(ordered_by_length)
    i = 0
    ordered_by_length = list(ordered_by_length.items())
    while i < num_storfs:
        pos_x, data_x = ordered_by_length[i]
        start_x = int(pos_x.split(',')[0])
        stop_x = int(pos_x.split(',')[-1])
        j = i+1
        while j < num_storfs:
            pos_y, data_y = ordered_by_length[j]
            start_y = int(pos_y.split(',')[0])
            stop_y = int(pos_y.split(',')[-1])
            if start_y >= stop_x or stop_y <= start_x:
                j+=1
                continue  # Not caught up yet / too far
            elif start_y >= start_x and stop_y <= stop_x:
                ordered_by_length.pop(j)
                num_storfs = len(ordered_by_length)
            else: # +1 needed for stop codon
                x = set(range(start_x,stop_x+1))
                y = set(range(start_y,stop_y+1))
                overlap = len(x.intersection(y))
                if overlap >= options.overlap_nt:
                    ordered_by_length.pop(j)
                    num_storfs = len(ordered_by_length)
                else:
                    j += 1
        num_storfs = len(ordered_by_length)
        i+=1
    filtered_storfs = OrderedDict(ordered_by_length)
    #### Clunky - reordering of StORFs - Through number found (pos then neg strand) or start position?
    if options.storf_order == 'start_pos':
        final_filtered_storfs = sortORFs(filtered_storfs)
    elif options.storf_order == 'strand':
        final_filtered_storfs = collections.OrderedDict()
        storf_nums = [item[-1] for item in filtered_storfs.values()]
        storf_nums = sorted(storf_nums)
        for num in storf_nums:
            for key,value in filtered_storfs.items():
                if value[-1] == num:
                    final_filtered_storfs.update({key:value})
                    break

    return final_filtered_storfs


def prepare_out(options, storfs, seq_id):
    gff_entries = []
    fasta_entries = {}
    for pos, data in storfs.items():
        sequence = data[0]
        strand = data[2]
        idx = data[5]
        pos_ = pos.split(',')
        start_stop = sequence[0:3]
        if len(pos_) == 3:
            mid = int(pos_[1]) - int(pos_[0])
            mid_stop = sequence[mid-3:mid]
        else:
            mid_stop = 'N/A'
        end_stop = sequence[-3:]
        start = int(pos_[0])
        stop = int(pos_[-1])
        ur_frame = int(data[1])
        storf_Type = data[4]
        native_seq = seq_id.replace('>','')
        ur_name = seq_id.replace('|', ':')
        length = len(sequence)
        if options.unannotated == True:
            if strand == '+':
                gff_start = str(start + 1 + int(ur_name.split('_')[-2])) # + 1 to adjust the first stop codon loci
                gff_stop = str(stop + int(ur_name.split('_')[-2]))
                if options.stop_inclusive == False:  # To remove the start and stop codon positions.
                    gff_start = gff_start + 3
                frame = (int(gff_stop) % 3) + 1
            elif strand == '-':
                gff_start = str(start - 2 + int(ur_name.split('_')[-2])) # -2 / -3 to adjust the first stop codon loci
                gff_stop = str(stop - 3 + int(ur_name.split('_')[-2]))
                if options.stop_inclusive == False:  # To remove the start and stop codon positions.
                    gff_stop = gff_stop - 3
                frame = (int(gff_stop) % 3) + 4

            storf_name = native_seq + '_' + storf_Type + '_' + str(idx) + ':' + gff_start + '-' + gff_stop

            gff_entries.append(native_seq.split('_UR')[0] + '\tStORF-Reporter\t' + options.feature_type + '\t' + gff_start + '\t' + gff_stop + '\t.\t' + data[2] +
                '\t.\tID=' + storf_name + ';UR=' + ur_name.replace('>','')  + ';UR_Stop_Locations=' + '-'.join(pos_) + ';Length=' + str(
                    length) + ';Strand=' + data[2] +
                ';Frame=' + str(frame) + ';UR_Frame=' + str(ur_frame) +
                ';Start_Stop=' + start_stop + ';Mid_Stop=' + mid_stop  + ';End_Stop=' + end_stop + ';StORF_Type=' + storf_Type + '\n')

            ### need to add - if con-storf then add middle stops
            fasta_entries.update({'>' + storf_name + ';UR=' + ur_name.replace('>','')  + ';UR_Stop_Locations=' + '-'.join(pos_) + ';Length=' +
                                str(length) + ';Strand=' + data[2] + ';Frame=' + str(frame) + ';UR_Frame=' + str(ur_frame) +
                ';Start_Stop=' + start_stop + ';End_Stop=' + end_stop + ';StORF_Type=' + storf_Type + '\n':data[0]})

#########################################################################
        elif options.intergenic == False: # Not done yet
            fa_id = (">" + str(storf_name) + "|" + str(start) + strand + str(stop) + "|Frame:" + str(
                frame) + '|Start_Stop=' + start_stop + '|Mid_Stop=' + mid_stop +
                     '|End_Stop=' + end_stop + '|StORF_Type:' + storf_Type + "\n")
            gff_entries.append(                native_seq + '\tStORF_Reporter\t' + options.feature_type + '\t' + gff_start + '\t' + gff_stop + '\t.\t' + data[2] +
                '\t.\tID=' + storf_name + ';UR=' + ur_name + ';UR_Stop_Locations=' + '-'.join(pos_) + ';Length=' + str(
                    length) +
                ';Frame=' + str(frame) + ';UR_Frame=' + str(ur_frame) +
                ';Start_Stop=' + start_stop + ';End_Stop=' + end_stop + ';StORF_Type=' + storf_Type + '\n')

    return gff_entries, fasta_entries


def write_gff(gff_entries,gff_out):
    ###GFF Out
    for entry in gff_entries:
            gff_out.write(entry)

def write_fasta(options, fasta_entries, fasta_out,aa_fasta_out):  # Some Lines commented out for BetaRun of ConStORF
    ###FASTA Prepare
    storf_num = 0 # This requires a much more elegant solution.
    ###FASTA Out
    for fasta_id, sequence in fasta_entries.items(): # could be made more efficient
        strand = fasta_id.split('Strand=')[1].split(';')
        if options.aa_only == False:# and options.translate == False:
            fasta_out.write(fasta_id)
            if options.line_wrap:
                wrapped = textwrap.wrap(sequence, width=60)
                for wrap in wrapped:
                    fasta_out.write(wrap + '\n')
            else:
                fasta_out.write(sequence + '\n')
        if options.translate == True or options.aa_only == True:
            aa_fasta_out.write(fasta_id)
            if "+" in strand:
                amino = translate_frame(sequence[0:])
                if options.stop_ident == False:
                    amino = amino.replace('*', '') # Remove * from sequences
                if options.line_wrap:
                    amino = textwrap.wrap(amino, width=60)
                    for wrap in amino:
                        aa_fasta_out.write(wrap + '\n')
                else:
                    aa_fasta_out.write(amino + '\n')
            if "-" in strand:
                amino = translate_frame(sequence[0:])
                if options.stop_ident == False:
                    amino = amino.replace('*', '') # Remove * from sequences
                if options.line_wrap:
                    amino = textwrap.wrap(amino, width=60)
                    for wrap in amino:
                        aa_fasta_out.write(wrap + '\n')
                else:
                    aa_fasta_out.write(amino + '\n')
        storf_num += 1


def find_storfs(working_frame,sequence_id,stops,sequence,storfs,short_storfs,con_StORFs,frames_covered,counter,lengths,strand,StORF_idx,short_StORF_idx,Con_StORF_idx,options):
    first = True
    con_StORF_tracker = ''
    next_stops = []
    start_stops = []
    seen_stops = []
    for stop in stops:  # Finds Stop-Stop#
        seen_stops.append(stop)
        if strand == '+':
            frame = (stop % 3) + 1
        elif strand == '-':
            frame = (stop % 3) + 4
        frames_covered.update({frame: 1})
        for next_stop in stops[counter + 1:]:
            length = abs((next_stop + 3) - stop)
            if length % 3 == 0 and next_stop not in seen_stops: # In frame and not already seen?
                    if length >= options.min_orf and length <= options.max_orf:
                        if not first and stop == prev_next_stop:
                            if prev_next_stop != con_StORF_tracker:
                                con_StORF_tracker = next_stop
                                seq = sequence[prev_stop:next_stop]
                                length = next_stop - prev_stop
                                ##### Needed to correct for negative frame loci
                                if working_frame == 'negative':
                                    rev_corrected_start, rev_corrected_mid, rev_corrected_stop = reverseCorrectLoci(sequence_id,prev_stop,stop,next_stop  + 3)
                                    con_StORF_Pos = ",".join([str(rev_corrected_start), str(rev_corrected_mid), str(rev_corrected_stop)])
                                else:
                                    con_StORF_Pos = ",".join([str(prev_stop), str(stop), str(next_stop  + 3)])
                                #####
                                con_length = abs((next_stop + 3) - prev_stop)
                                con_StORFs.update({con_StORF_Pos: [seq, str(frame), strand, con_length,'Con-StORF',Con_StORF_idx]})
                                Con_StORF_idx +=1
                            elif stop == con_StORF_tracker:
                                con_StORF_tracker = next_stop
                                prev_con_StORF = next(reversed(con_StORFs.keys())) #Get last key
                                seq_start = int(prev_con_StORF.split(',')[0])
                                seq = sequence[seq_start:next_stop  + 3]
                                length = next_stop - seq_start # Check
                                prev_con_StORF = prev_con_StORF_CHECKER(prev_con_StORF,sequence,options)
                                ##### Needed to correct for negative frame loci
                                if working_frame == 'negative':
                                    rev_corrected_start, rev_corrected_mid, rev_corrected_stop = reverseCorrectLoci(sequence_id,prev_stop,stop,next_stop  + 3)
                                    con_StORF_Pos = ",".join([str(rev_corrected_start), str(rev_corrected_mid), str(rev_corrected_stop)])
                                else:
                                    con_StORF_Pos = ",".join([str(prev_stop), str(stop), str(next_stop  + 3)])
                                #####
                                if options.olap_filtering == 'both-strand':
                                    con_StORFs.popitem()
                                con_length = abs((next_stop + 3) - prev_stop)
                                con_StORFs.update({con_StORF_Pos: [seq, str(frame), strand, con_length,'Con-StORF',Con_StORF_idx]})
                                Con_StORF_idx +=1

                        if options.olap_filtering == 'none': # This could be made more efficient
                            seq = sequence[stop:next_stop  + 3]
                            ##### Needed to correct for negative frame loci
                            if working_frame == 'negative':
                                rev_corrected_start, rev_corrected_stop = reverseCorrectLoci(sequence_id,stop,None,next_stop  + 3)
                                storfs.update({",".join([str(rev_corrected_start), str(rev_corrected_stop)]): [seq, str(frame), strand, length, 'StORF', StORF_idx]})
                            else:
                                storfs.update({",".join([str(stop), str(next_stop  + 3)]): [seq, str(frame), strand, length,'StORF',StORF_idx]})
                            #####
                            StORF_idx +=1
                            lengths.append(length)
                        elif not first:
                            if stop > prev_stop and next_stop < prev_next_stop: # Check
                                break
                            storf = set(range(stop, next_stop + 4)) # + 4 to account for set use
                            storf_overlap = len(prev_storf.intersection(storf))
                            if storf_overlap <= options.overlap_nt:
                                seq = sequence[stop:next_stop  + 3]
                                ##### Needed to correct for negative frame loci
                                if working_frame == 'negative':
                                    rev_corrected_start, rev_corrected_stop = reverseCorrectLoci(sequence_id,stop,None,next_stop  + 3)
                                    storfs.update({",".join([str(rev_corrected_start), str(rev_corrected_stop)]): [
                                        seq, str(frame), strand, length, 'StORF', StORF_idx]})
                                else:
                                    storfs.update({",".join([str(stop), str(next_stop  + 3)]): [seq, str(frame), strand,
                                                                                               length, 'StORF',
                                                                                               StORF_idx]})
                                #####
                                StORF_idx +=1
                                seen_stops.append(next_stop)#  + 3)
                                prev_storf = storf
                                prev_stop = stop
                                prev_next_stop = next_stop # Check
                                next_stops.append(next_stop)
                                start_stops.append(stop)
                                prevlength = prev_next_stop - prev_stop
                            elif storf_overlap >= options.overlap_nt and options.olap_filtering == 'both-strand':  # and length > prevlength:
                                if length > prevlength:
                                    storfs.popitem()
                                    seq = sequence[stop:next_stop  + 3]
                                    ##### Needed to correct for negative frame loci
                                    if working_frame == 'negative':
                                        rev_corrected_start, rev_corrected_stop = reverseCorrectLoci(sequence_id,stop,None,next_stop  + 3)
                                        storfs.update({",".join(
                                            [str(rev_corrected_start), str(rev_corrected_stop)]): [seq, str(frame),
                                                                                                       strand, length,
                                                                                                       'StORF',
                                                                                                       StORF_idx]})
                                    else:
                                        storfs.update({",".join([str(stop), str(next_stop  + 3)]): [seq, str(frame),
                                                                                                   strand, length,
                                                                                                   'StORF', StORF_idx]})
                                    #####
                                    StORF_idx +=1
                                    seen_stops.append(next_stop)
                                    prev_storf = storf
                                    prev_stop = stop
                                    prev_next_stop = next_stop
                                    next_stops.append(next_stop)
                                    start_stops.append(stop)
                                    prevlength = prev_next_stop - prev_stop
                            else: # If filtering is none or single-strand, we do not remove overlapping StORFs on the same strand
                                seq = sequence[stop:next_stop  + 3]
                                ##### Needed to correct for negative frame loci
                                if working_frame == 'negative':
                                    rev_corrected_start, rev_corrected_stop = reverseCorrectLoci(sequence_id,stop,None,next_stop  + 3)
                                    storfs.update({",".join([str(rev_corrected_start), str(rev_corrected_stop)]): [
                                        seq, str(frame), strand, length, 'StORF', StORF_idx]})
                                else:
                                    storfs.update({",".join([str(stop), str(next_stop  + 3)]): [seq, str(frame), strand,
                                                                                               length, 'StORF',
                                                                                               StORF_idx]})
                                #####
                                StORF_idx += 1
                                seen_stops.append(next_stop)
                                prev_storf = storf
                                prev_stop = stop
                                prev_next_stop = next_stop
                                next_stops.append(next_stop)
                                start_stops.append(stop)
                                prevlength = prev_next_stop - prev_stop
                        elif first:
                            if options.partial_storf: # upstream partial StORF_Reporter
                                if stop > options.min_orf and frames_covered[frame] != 1:
                                    seq = sequence[0:stop + 3] #Start of seq to first stop identified
                                    #ps_seq = cut_seq(seq, '-')
                                    ##### Needed to correct for negative frame loci
                                    if working_frame == 'negative':
                                        rev_corrected_start, rev_corrected_stop = reverseCorrectLoci(sequence_id,stop,None,next_stop  + 3)
                                        storfs.update({",".join(
                                            [str(rev_corrected_start), str(rev_corrected_stop)]): [seq, str(frame),
                                                                                                       strand, length,
                                                                                                       'StORF',
                                                                                                       StORF_idx]})
                                    else:
                                        storfs.update({",".join([str(stop), str(next_stop  + 3)]): [seq, str(frame),
                                                                                                   strand, length,
                                                                                                   'StORF', StORF_idx]})
                                    #####
                                    StORF_idx +=1
                            seq = sequence[stop:next_stop  + 3]
                            length = next_stop - stop
                            ##### Needed to correct for negative frame loci
                            if working_frame == 'negative':
                                rev_corrected_start, rev_corrected_stop = reverseCorrectLoci(sequence_id,stop,None,next_stop  + 3)
                                storfs.update({",".join([str(rev_corrected_start), str(rev_corrected_stop)]): [seq, str(frame), strand, length, 'StORF', StORF_idx]})
                            else:
                                storfs.update({",".join([str(stop), str(next_stop  + 3)]): [seq, str(frame), strand, length,'StORF',StORF_idx]})
                            #####
                            StORF_idx +=1
                            seen_stops.append(next_stop)
                            prev_storf = set(range(stop, next_stop + 4)) # + 4 to account for set use
                            prev_stop = stop
                            prev_next_stop = next_stop
                            next_stops.append(next_stop)
                            start_stops.append(stop)
                            prevlength = prev_next_stop - prev_stop
                            first = False
                        break

                    if options.short_storfs != False and length >= 30: # Report short (<= 120) StORFs with less than 50% overlap with other larger StORFs
                        seq = sequence[stop:next_stop  + 3]
                        length = next_stop - stop # Check
                        short_storfs.update({",".join([str(stop), str(next_stop  + 3)]): [seq, str(frame), strand, length,'Short-StORF', short_StORF_idx]})
                        short_StORF_idx +1
                    else:
                        break
        counter +=1
    if options.partial_storf:  # downstream partial StORF_Reporter - Last Stop to end of sequence
        try:
            if (len(sequence) - stop) > options.min_orf:
                seq = sequence[stop:len(sequence)]  # Start of seq to first stop identified - not working
                ps_seq = cut_seq(seq, '+')
                storfs.update({",".join([str(stop), str(len(sequence))]): [ps_seq, str(frame), strand, stop,'Partial-StORF',StORF_idx]})
                StORF_idx +=1
        except UnboundLocalError:
            pass
    return storfs, short_storfs, con_StORFs, frames_covered, counter, lengths, StORF_idx, Con_StORF_idx

def STORF_Finder(options, sequence_info, sequence_id, fasta_out, aa_fasta_out, gff_out): #Main Function
    sequence_region_length = sequence_info[0]
    sequence = sequence_info[1]
    stops = []
    frames_covered = OrderedDict()
    for x in range (1,7):
        frames_covered.update({x: 0})
    for stop_codon in options.stop_codons.split(','): #Find all Stops in seq
        stops += [match.start() for match in re.finditer(re.escape(stop_codon), sequence)]
    stops.sort()
    storfs = OrderedDict()
    short_storfs = OrderedDict()
    con_StORFs = OrderedDict()
    counter = 0
    lengths = []
    StORF_idx = 0
    short_StORF_idx = 0
    Con_StORF_idx = 0
    storfs, short_storfs, con_StORFs,frames_covered,counter,lengths,StORF_idx,Con_StORF_idx = find_storfs("positive",sequence_id,stops,sequence,storfs,short_storfs,con_StORFs,frames_covered,counter,lengths,'+',StORF_idx,short_StORF_idx,Con_StORF_idx,options)
    ###### Reversed Comppliment
    sequence_rev = revCompIterative(sequence)
    stops = []
    for stop_codon in options.stop_codons.split(','): #Find all Stops in seq
        stops += [match.start() for match in re.finditer(re.escape(stop_codon), sequence_rev)]
    stops.sort()
    counter = 0
    storfs, short_storfs, con_StORFs,frames_covered,counter,lengths,StORF_idx,Con_StORF_idx = find_storfs("negative",sequence_id,stops,sequence_rev,storfs,short_storfs,con_StORFs,frames_covered,counter,lengths,'-',StORF_idx,short_StORF_idx,Con_StORF_idx,options)
    #Potential run-through StORFs
    if options.whole_contig:
        for frame,present in frames_covered.items():
            if present == 0:
                if frame <4:
                    wc_seq = sequence[frame-1:]
                    wc_seq = cut_seq(wc_seq,'+')
                    storfs.update({",".join([str(0), str(len(sequence))]): [wc_seq, str(frame), '+', len(sequence),'Run-Through-StORF',StORF_idx]})
                    StORF_idx +=1
                else:
                    wc_seq = sequence_rev[frame-4:]
                    wc_seq = cut_seq(wc_seq,'+')
                    storfs.update({",".join([str(0), str(len(sequence))]): [wc_seq, str(frame), '-', len(sequence_rev),'Run-Through-StORF',StORF_idx]})
                    StORF_idx +=1

####################################### Writing output
    ######## Only StORFs
    #Check if there are StORFs to report
    if options.con_storfs == False and options.con_only == False and options.short_storfs == False:
        if bool(storfs):
            if options.start_filtering == True:
                storfs = start_filtering(storfs)
            if options.olap_filtering == 'both-strand':
                storfs = tile_filtering(storfs, options)  # Filtering
            storfs = OrderedDict(sorted(storfs.items(), key=lambda e: tuple(map(int, e[0].split(",")))))  # Reorder by start position
            if options.reporter == True:
                return storfs
            ###Data Prepare
            gff_entries, fasta_entries = prepare_out(options, storfs, sequence_id)
            write_fasta(options, fasta_entries, fasta_out, aa_fasta_out)
            if not options.aa_only:
                write_gff(gff_entries, gff_out)

    ###### Only Short-StORFs
    elif options.short_storfs != False and options.short_storfs_only == True: # Short-StORFs ONLY
        if bool(short_storfs):
            if options.start_filtering == True:
                short_storfs = start_filtering(short_storfs)
            if options.olap_filtering == 'both-strand':
                short_storfs = tile_filtering(short_storfs, options)  # Filtering
            short_storfs = OrderedDict(sorted(short_storfs.items(), key=lambda e: tuple(map(int, e[0].split(",")))))  # Reorder by start position
            ###Data Prepare
            gff_entries, fasta_entries = prepare_out(options, short_storfs, sequence_id)
            write_fasta(options, fasta_entries, fasta_out, aa_fasta_out)
            if not options.aa_only:
                write_gff(gff_entries, gff_out)

    ### StORFs and Short-StORFs
    elif options.short_storfs != False and options.short_storfs_only == False:
        ### Don't allow short-storfs to overlap with storfs
        if options.short_storfs == 'Nolap':
            all_StORFs = {**storfs, **short_storfs}
            if options.olap_filtering == 'both-strand':
                all_StORFs = tile_filtering(all_StORFs,options) # Filtering
            all_StORFs = OrderedDict(sorted(all_StORFs.items(), key=lambda e: tuple(map(int, e[0].split(","))))) # Reorder by start position
            if options.reporter == True:
                return all_StORFs
        ### short-storfs can onverlap with storfs
        elif options.short_storfs == 'Olap':
            if options.olap_filtering == 'both-strand': # Filter individually
                storfs = tile_filtering(storfs,options) # Filtering
                short_storfs = tile_filtering(short_storfs,options) # Filtering
            all_StORFs = {**storfs, **short_storfs}
            all_StORFs = OrderedDict(sorted(all_StORFs.items(), key=lambda e: tuple(map(int, e[0].split(",")))))
            ###Data Prepare
        gff_entries, fasta_entries = prepare_out(options, all_StORFs, sequence_id)
        write_fasta(options, fasta_entries, fasta_out, aa_fasta_out)
        if not options.aa_only:
            write_gff(gff_entries, gff_out)


    ####### StORFs and Con-StORFs
    elif options.con_storfs == True and options.con_only == False:
        all_StORFs = {**storfs, **con_StORFs}
        if bool(storfs):
            if options.olap_filtering == 'both-strand':
                all_StORFs = tile_filtering(all_StORFs, options) # Filtering
            all_StORFs = OrderedDict(sorted(all_StORFs.items(), key=lambda e: tuple(map(int, e[0].split(","))))) # Reorder by start position
            if options.reporter == True:
                return all_StORFs
            ###Data Prepare
            gff_entries, fasta_entries = prepare_out(options, all_StORFs, sequence_id)
            write_fasta(options, fasta_entries, fasta_out, aa_fasta_out)
            if not options.aa_only:
                write_gff(gff_entries, gff_out)

    ###### Con-StORFs only
    elif options.con_only == True:
        if options.olap_filtering == 'both-strand':
            con_StORFs = tile_filtering(con_StORFs, options)
            con_StORFs = OrderedDict(sorted(con_StORFs.items(), key=lambda e: tuple(map(int, e[0].split(","))))) # Reorder by start position
        else:
            con_StORFs = OrderedDict(sorted(con_StORFs.items(), key=lambda e: tuple(map(int, e[0].split(","))))) # Reorder by start position
        if options.reporter == True:
            return con_StORFs
        ###Data Prepare
        gff_entries, fasta_entries = prepare_out(options, con_StORFs, sequence_id)
        write_fasta(options, fasta_entries, fasta_out, aa_fasta_out)
        if not options.aa_only:
            write_gff(gff_entries, gff_out)
    ###### Below won't work..?
    elif options.verbose == True:
        print("No StOFS Found")

def fasta_load(fasta_in, sequence_regions, sequences):
    first = True
    sequence_region_length = 0
    for line in fasta_in:
        line = line.strip()
        if '##sequence-region' in line:
            sequence_region_length = int(line.split(' ')[-1]) # bug and wont work on non-UR runs
            sequence_regions.append(line)
        elif line.startswith((';','\n','#')):
            continue
        elif line.startswith('>') and not first:
            sequences.update({sequence_name: [sequence_region_length,seq.strip()]})
            seq = ''
            sequence_name = line.strip()
        elif line.startswith('>'):
            seq = ''
            sequence_name = line.strip()
        elif line:
            seq += str(line)
            first = False
    sequences.update({sequence_name: [sequence_region_length,seq.strip()]})
    return sequence_regions, sequences

## Function to control how StORF-Finder handles StORF-Reporter output
def StORF_Reported(options, Contigs):
    Reporter_StORFs = collections.OrderedDict()
    for Contig_ID, Contig_URs in Contigs.items():
        Reporter_StORFs.update({Contig_ID:[]})
        URs = Contig_URs[3]
        try:
            for UR in URs:
                if len(URs[UR][1]) >= options.min_orf: # Here
                    contig_length = Contigs[Contig_ID][1]
                    sequence_info = [contig_length,URs[UR][1]]
                    if UR.split('_')[0] == '0': # This is to account for the GFF base-1 system
                        sequence_id = "1_" + UR.split('_')[1]
                    else:
                        sequence_id = UR # mockup sequence_id in correct format for later
                    StORFs = STORF_Finder(options, sequence_info, sequence_id, None, None, None) # need to pass seq and original seq length
                    if StORFs: #  Left out for now to allow for tracking of non-StORF URs
                        for StORF in StORFs.values():
                            StORF.append(URs[UR][0]) # True UR
                            StORF.append(sequence_id) # Extended UR
                        Reporter_StORFs[Contig_ID].append(StORFs)
        except TypeError:
            if options.verbose == True:
                print("No URs in seq")
    return Reporter_StORFs


def main():
    print("Thank you for using StORF-Reporter -- A detailed user manual can be found at https://github.com/NickJD/StORF-Reporter\nPlease report any issues to: https://github.com/NickJD/StORF-Reporter/issues\n#####")

    parser = argparse.ArgumentParser(description='StORF-Reporter ' + StORF_Reporter_Version + ': StORF-Finder Run Parameters.')
    parser._action_groups.pop()

    required = parser.add_argument_group('Required Arguments')
    required.add_argument('-f', action="store", dest='fasta', required=True,
                        help='Input FASTA File - (UR_Extractor output)')

    optional = parser.add_argument_group('Optional Arguments')
    optional.add_argument('-reporter', action="store", dest='reporter', default=False, required=False,
                        help=argparse.SUPPRESS)
    optional.add_argument('-ua', dest='unannotated', action='store', default=True, type=eval, choices=[True, False],
                        help='Default - Treat input as Unannotated: Use "-ua False" for standard fasta')
    optional.add_argument('-wc', action="store", dest='whole_contig', default=False, type=eval, choices=[True, False],
                        help='Default - False: StORFs reported across entire sequence')
    optional.add_argument('-ps', action="store", dest='partial_storf', default=False, type=eval, choices=[True, False],
                        help='Default - False: Partial StORFs reported')
    optional.add_argument('-olap_filt', action='store', dest='olap_filtering', default='both-strand', const='both-strand', nargs='?',
                        choices=['none', 'single-strand', 'both-strand'],
                        help='Default - "both-strand": Filtering level "none" is not recommended, "single-strand" for single strand filtering '
                             'and both-strand for both-strand longest-first tiling')
    optional.add_argument('-start_filt', action="store", dest='start_filtering', default=False, type=eval, choices=[True, False],
                        help='Default - False: Filter out StORFs without at least one of the 3 common start codons (best used for short-storfs).')
    optional.add_argument('-con_storfs', action="store", dest='con_storfs', default=False, type=eval, choices=[True, False],
                        help='Default - False: Output Consecutive StORFs')
    optional.add_argument('-con_only', action="store", dest='con_only', default=False, type=eval, choices=[True, False],
                        help='Default - False: Only output Consecutive StORFs')
    optional.add_argument('-short_storfs', action="store", dest='short_storfs', default=False, type=str, choices=[False, 'Nolap', 'Olap'],
                        help='Default - False: Run StORF-Finder in "Short-StORF" mode. Will only return StORFs between 30 and 120 nt '
                             'that do not overlap longer StORFs - Only works with StORFs for now. "Nolap" will filter Short-StORFs which are'
                             'overlapped by StORFs and Olap will report Short-StORFs which do overlap StORFs. Overlap is defined by "-olap".')
    optional.add_argument('-short_storfs_only', action="store", dest='short_storfs_only', default=False, type=eval, choices=[True, False],
                        help='Default - True. Only report Short-StORFs?')
    optional.add_argument('-stop_ident', action="store", dest='stop_ident', default=False, choices=[True, False],
                        help='Default - True: Identify Stop Codon positions with \'*\'')
    optional.add_argument('-f_type', action='store', dest='feature_type', default='CDS', const='StORF', nargs='?',
                        choices=['StORF', 'CDS', 'ORF'],
                        help='Default - "StORF": Which GFF feature type for StORFs to be reported as in GFF')
    optional.add_argument('-minorf', action="store", dest='min_orf', default=99, type=int,
                        help='Default - 99: Minimum StORF size in nt')
    optional.add_argument('-maxorf', action="store", dest='max_orf', default=60000, type=int,
                        help='Default - 60kb: Maximum StORF size in nt')
    optional.add_argument('-codons', action="store", dest='stop_codons', default="TAG,TGA,TAA",
                        help='Default - (\'TAG,TGA,TAA\'): List Stop Codons to use')
    optional.add_argument('-olap', action="store", dest='overlap_nt', default=50, type=int,
                        help='Default - 50: Maximum number of nt of a StORF which can overlap another StORF.')
    optional.add_argument('-s', action="store", dest='suffix', required=False,
                        help='Default - Do not append suffix to genome ID')
    optional.add_argument('-so', action="store", dest='storf_order', default='start_pos', nargs='?', choices=['start_pos','strand'],
                        required=False,
                        help='Default - Start Position: How should StORFs be ordered when >1 reported in a single UR.')
    optional.add_argument('-spos', action="store", dest='stop_inclusive', default=True, type=eval, choices=[True, False],
                        help='Default - False: Print out StORF positions inclusive of first stop codon')


    output = parser.add_argument_group('Output')
    output.add_argument('-oname', action="store", dest='o_name', required=False,
                        help='Default - Appends \'_StORF-R\' to end of input FASTA filename')
    output.add_argument('-odir', action="store", dest='o_dir', required=False,
                        help='Default -  Same directory as input FASTA')
    # output.add_argument('-af', action="store", dest='affix', required=False,
    #                     help='Default - \'StORF-R\': can be used to append an identifier to output filename '
    #                          'to distinguish Con-StORF from StORF runs)')
    output.add_argument('-gff', action='store', dest='gff', default=True, type=eval, choices=[True, False],
                        help='Default - True: Output a GFF file')
    output.add_argument('-aa', action="store", dest='translate', default=False, type=eval, choices=[True, False],
                        help='Default - False: Report StORFs as amino acid sequences')
    output.add_argument('-aa_only', action="store", dest='aa_only', default=False, type=eval, choices=[True, False],
                        help='Default - False: Only output Amino Acid Fasta')
    output.add_argument('-lw', action="store", dest='line_wrap', default=True, type=eval, choices=[True, False],
                        help='Default - True: Line wrap FASTA sequence output at 60 chars')
    output.add_argument('-gff_fasta', action="store", dest='gff_fasta', default=False, type=eval, choices=[True, False],
                        help='Default - False: Report all gene sequences (nt) at the bottom of GFF files in Prokka output mode')
    output.add_argument('-gz', action='store', dest='gz', default='False', type=eval, choices=[True, False],
                        help='Default - False: Output as .gz')
    optional.add_argument('-nout', action='store', dest='nout', default='False', type=eval, choices=[True, False],
                        help=argparse.SUPPRESS)

    misc = parser.add_argument_group('Misc')
    misc.add_argument('-v', action='store', dest='verbose', default='False', type=eval, choices=[True, False],
                        help='Default - False: Print out runtime status')

    options = parser.parse_args()


    #ns_nt = defaultdict  # Used to Record non-standard nucleotides - not implemented yet
    ##### Load in fasta file
    sequences = OrderedDict()
    sequence_regions = []
    if options.reporter == False:
        try: # Detect whether fasta files are .gz or text and read accordingly
            fasta_in = gzip.open(options.fasta,'rt')
            sequence_regions, sequences = fasta_load(fasta_in, sequence_regions, sequences)
        except:
            fasta_in = open(options.fasta,'r')
            sequence_regions, sequences =  fasta_load(fasta_in, sequence_regions, sequences)
        if options.verbose == True:
            print(fasta_in.name)
    #### Output Directory and Filename handling
    if options.o_dir == None and options.o_name == None:
        tmp_extension = options.fasta.split('.')[-1]  # could be .fa/.fasta etc
        output_file = options.fasta.replace('.' + tmp_extension, '')
        output_file = output_file + '_StORF-R'
    elif options.o_dir != None and options.o_name != None:
        output_file = options.o_dir
        output_file = output_file + options.o_name
    elif options.o_dir != None:
        tmp_extension = options.fasta.split('.')[-1]  # could be .fa/.fasta etc
        output_file = options.fasta.replace('.' + tmp_extension, '').split('/')[-1]
        output_file = options.o_dir + output_file + '_StORF-R'
    elif options.o_name != None:
        tmp_filename = options.fasta.split('/')[-1]  # could be .fa/.fasta etc
        output_file = options.fasta.replace(tmp_filename, '')
        output_file = output_file + options.o_name

    if not options.gz: # Clear fasta and gff files if not empty - Needs an elegant solution
        if not options.aa_only:
            gff_out = open(output_file + '.gff', 'w', newline='\n', encoding='utf-8')
            gff_out.write("##gff-version\t3\n#\tStORF-Reporter - Stop ORF Predictions\n#\tRun Date:" + str(date.today()) + '\n')
            gff_out.write('##StORF-Reporter ' + StORF_Reporter_Version + '\n')
            for seq_reg in sequence_regions:
                gff_out.write(seq_reg + '\n')
            gff_out.write("##Original File: " + options.fasta.split('/')[-1] + '\n\n')
            fasta_out = open(output_file + '.fasta', 'w', newline='\n', encoding='utf-8')
            if options.translate:
                aa_fasta_out = open(output_file + '_aa.fasta', 'w', newline='\n', encoding='utf-8')
            else:
                aa_fasta_out = None
        elif options.aa_only:
            aa_fasta_out = open(output_file + '_aa.fasta', 'w', newline='\n', encoding='utf-8')
    elif options.gz:
        if not options.aa_only:
            gff_out = gzip.open(output_file + '.gff.gz', 'wt', newline='\n', encoding='utf-8')
            gff_out.write("##gff-version\t3\n#\tStORF-Reporter - Stop ORF Predictions\n#\tRun Date:" + str(date.today()) + '\n')
            gff_out.write('##StORF-Reporter ' + StORF_Reporter_Version + '\n')
            for seq_reg in sequence_regions:
                gff_out.write(seq_reg + '\n')
            gff_out.write("##Original File: " + options.fasta + '\n\n')
            fasta_out = gzip.open(output_file + '.fasta.gz', 'wt', newline='\n', encoding='utf-8')
            if options.translate:
                aa_fasta_out = gzip.open(output_file + '_aa.fasta.gz', 'wt', newline='\n', encoding='utf-8')
            else:
                aa_fasta_out = None
        elif options.aa_only:
            aa_fasta_out = gzip.open(output_file + '_aa.fasta.gz', 'wt', newline='\n', encoding='utf-8')

    for sequence_id, sequence_info in sequences.items():
        if len(sequence_info[1]) >= options.min_orf:
            STORF_Finder(options, sequence_info, sequence_id, fasta_out, aa_fasta_out, gff_out)

if __name__ == "__main__":
    main()
    print("Complete")
